import numpy as np
import pytest
from sklearn.preprocessing import MaxAbsScaler

from darts.dataprocessing.transformers import Scaler
from darts.datasets import AirPassengersDataset
from darts.tests.conftest import RAY_AVAILABLE, TORCH_AVAILABLE, tfm_kwargs

if not RAY_AVAILABLE:
    pytest.skip(
        f"Ray not available. {__name__} tests will be skipped.",
        allow_module_level=True,
    )

from ray import tune
from ray.tune.tuner import Tuner

if TORCH_AVAILABLE:
    from pytorch_lightning.callbacks import Callback, EarlyStopping
    from ray.tune.integration.pytorch_lightning import TuneReportCheckpointCallback
    from torchmetrics import (
        MeanAbsoluteError,
        MeanAbsolutePercentageError,
        MetricCollection,
    )

    from darts.models import NBEATSModel


class TestRay:
    series = AirPassengersDataset().load().astype(np.float32)

    val_length = 36
    train, val = series.split_after(val_length)

    # scale
    scaler = Scaler(MaxAbsScaler())
    train = scaler.fit_transform(train)
    val = scaler.transform(val)

    @pytest.mark.skipif(not TORCH_AVAILABLE, reason="requires torch")
    def test_ray_torch_model(self, tmpdir_fn):
        """Check that ray works as expected with a torch-based model"""

        def train_model(model_args, callbacks, train, val):
            torch_metrics = MetricCollection([
                MeanAbsolutePercentageError(),
                MeanAbsoluteError(),
            ])

            # Create the model using model_args from Ray Tune
            model = NBEATSModel(
                input_chunk_length=4,
                output_chunk_length=3,
                n_epochs=2,
                torch_metrics=torch_metrics,
                pl_trainer_kwargs={
                    **tfm_kwargs["pl_trainer_kwargs"],
                    "callbacks": callbacks,
                },
                **model_args,
            )

            model.fit(
                series=train,
                val_series=val,
            )

        # Early stop callback
        my_stopper = EarlyStopping(
            monitor="val_MeanAbsolutePercentageError",
            patience=5,
            min_delta=0.05,
            mode="min",
        )

        # set up ray tune callback
        class TuneReportCallback(TuneReportCheckpointCallback, Callback):
            pass

        tune_callback = TuneReportCallback(
            {
                "loss": "val_loss",
                "MAPE": "val_MeanAbsolutePercentageError",
            },
            on="validation_end",
        )

        # Define the trainable function that will be tuned by Ray Tune
        train_fn_with_parameters = tune.with_parameters(
            train_model,
            callbacks=[tune_callback, my_stopper],
            train=self.train,
            val=self.val,
        )

        # define the hyperparameter space
        param_space = {
            "batch_size": tune.choice([8, 16]),
            "num_blocks": tune.choice([1, 2]),
            "num_stacks": tune.choice([2, 4]),
        }

        # the number of combinations to try
        num_samples = 2

        # Create the Tuner object and run the hyperparameter search
        tuner = Tuner(
            trainable=train_fn_with_parameters,
            param_space=param_space,
            tune_config=tune.TuneConfig(
                metric="MAPE", mode="min", num_samples=num_samples
            ),
            run_config=tune.RunConfig(name="tune_darts"),
        )
        tuner.fit()



"""
Forecasting Model Explainer Base Class

A `_ForecastingModelExplainer` takes a fitted forecasting model as input and generates explanations for it.
"""

from abc import ABC, abstractmethod
from collections.abc import Sequence
from typing import Optional, Union

from darts import TimeSeries
from darts.explainability.explainability_result import _ExplainabilityResult
from darts.explainability.utils import process_horizons_and_targets, process_input
from darts.logging import get_logger, raise_log
from darts.models.forecasting.forecasting_model import ForecastingModel

logger = get_logger(__name__)

MIN_BACKGROUND_SAMPLE = 10


class _ForecastingModelExplainer(ABC):
    @abstractmethod
    def __init__(
        self,
        model: ForecastingModel,
        background_series: Optional[Union[TimeSeries, Sequence[TimeSeries]]] = None,
        background_past_covariates: Optional[
            Union[TimeSeries, Sequence[TimeSeries]]
        ] = None,
        background_future_covariates: Optional[
            Union[TimeSeries, Sequence[TimeSeries]]
        ] = None,
        requires_background: bool = False,
        requires_covariates_encoding: bool = False,
        check_component_names: bool = False,
        test_stationarity: bool = False,
    ):
        """
        The base class for forecasting model explainers. It defines the *minimal* behavior that all
        forecasting model explainers support.

        Naming:

        - A background series is a `TimeSeries` with which to 'train' the `Explainer` model.
        - A foreground series is the `TimeSeries` to explain using the fitted `Explainer` model.

        Parameters
        ----------
        model
            A `ForecastingModel` to be explained. It must be fitted first.
        background_series
            A series or list of series to *train* the `_ForecastingModelExplainer` along with any foreground series.
            Consider using a reduced well-chosen background to reduce computation time.
            Optional if `model` was fit on a single target series. By default, it is the `series` used at fitting time.
            Mandatory if `model` was fit on multiple (sequence of) target series.
        background_past_covariates
            A past covariates series or list of series that the model needs once fitted.
        background_future_covariates
            A future covariates series or list of series that the model needs once fitted.
        requires_background
            Whether the explainer requires background series as an input. If `True`, raises an error if no background
            series were provided and `model` was fit using multiple series.
        requires_covariates_encoding
            Whether to apply the model's encoders to the input covariates. This should only be `True` if the
            Explainer will not call model methods `fit()` or `predict()` directly.
        check_component_names
            Whether to enforce that, in the case of multiple time series, all series of the same type (target and/or
            *_covariates) must have the same component names.
        test_stationarity
            Whether to raise a warning if not all `background_series` are stationary.
        """
        if not model._fit_called:
            raise_log(
                ValueError(
                    f"The model must be fitted before instantiating a {self.__class__.__name__}."
                ),
                logger,
            )
        self.model = model
        # default forecasting horizon
        self.n: Optional[int] = getattr(self.model, "output_chunk_length", None)

        # check background input validity and process it
        (
            self.background_series,
            self.background_past_covariates,
            self.background_future_covariates,
            self.target_components,
            self.static_covariates_components,
            self.past_covariates_components,
            self.future_covariates_components,
        ) = process_input(
            model=model,
            input_type="background",
            series=background_series,
            past_covariates=background_past_covariates,
            future_covariates=background_future_covariates,
            fallback_series=model.training_series,
            fallback_past_covariates=model.past_covariate_series,
            fallback_future_covariates=model.future_covariate_series,
            check_component_names=check_component_names,
            requires_input=requires_background,
            requires_covariates_encoding=requires_covariates_encoding,
            test_stationarity=test_stationarity,
        )
        self.requires_foreground = self.background_series is None
        self.requires_covariates_encoding = requires_covariates_encoding
        self.check_component_names = check_component_names
        self.test_stationarity = test_stationarity

    @abstractmethod
    def explain(
        self,
        foreground_series: Optional[Union[TimeSeries, Sequence[TimeSeries]]] = None,
        foreground_past_covariates: Optional[
            Union[TimeSeries, Sequence[TimeSeries]]
        ] = None,
        foreground_future_covariates: Optional[
            Union[TimeSeries, Sequence[TimeSeries]]
        ] = None,
        horizons: Optional[Sequence[int]] = None,
        target_components: Optional[Sequence[str]] = None,
    ) -> _ExplainabilityResult:
        """
        Explains a foreground time series, and returns a :class:`_ExplainabilityResult
        <darts.explainability.explainability_result._ExplainabilityResult>` that can be used for downstream tasks.

        Parameters
        ----------
        foreground_series
            Optionally, one or a sequence of target `TimeSeries` to be explained. Can be multivariate.
            If not provided, the background `TimeSeries` will be explained instead.
        foreground_past_covariates
            Optionally, one or a sequence of past covariates `TimeSeries` if required by the forecasting model.
        foreground_future_covariates
            Optionally, one or a sequence of future covariates `TimeSeries` if required by the forecasting model.
        horizons
            Optionally, an integer or sequence of integers representing the future time steps to be explained.
            `1` corresponds to the first timestamp being forecasted.
            All values must be `<=output_chunk_length` of the explained forecasting model.
        target_components
            Optionally, a string or sequence of strings with the target components to explain.

        Returns
        -------
        _ExplainabilityResult
            The explainability result.
        """
        pass

    def _process_foreground(
        self,
        foreground_series: Optional[Union[TimeSeries, Sequence[TimeSeries]]] = None,
        foreground_past_covariates: Optional[
            Union[TimeSeries, Sequence[TimeSeries]]
        ] = None,
        foreground_future_covariates: Optional[
            Union[TimeSeries, Sequence[TimeSeries]]
        ] = None,
    ):
        return process_input(
            model=self.model,
            input_type="foreground",
            series=foreground_series,
            past_covariates=foreground_past_covariates,
            future_covariates=foreground_future_covariates,
            fallback_series=self.background_series,
            fallback_past_covariates=self.background_past_covariates,
            fallback_future_covariates=self.background_future_covariates,
            check_component_names=self.check_component_names,
            requires_input=self.requires_foreground,
            requires_covariates_encoding=self.requires_covariates_encoding,
            test_stationarity=self.test_stationarity,
        )

    def _process_horizons_and_targets(
        self,
        horizons: Optional[Union[int, Sequence[int]]],
        target_components: Optional[Union[str, Sequence[str]]],
    ) -> tuple[Sequence[int], Sequence[str]]:
        return process_horizons_and_targets(
            horizons=horizons,
            fallback_horizon=self.n,
            target_components=target_components,
            fallback_target_components=self.target_components,
            check_component_names=self.check_component_names,
        )



import inspect
from collections.abc import Sequence
from typing import Literal, Optional, Union

import numpy as np
import pandas as pd

from darts.logging import get_logger
from darts.timeseries import TimeSeries
from darts.utils.historical_forecasts.utils import (
    _get_historical_forecast_boundaries,
    _process_predict_start_points_bounds,
)
from darts.utils.utils import generate_index

logger = get_logger(__name__)


def _optimized_historical_forecasts(
    model,
    series: Sequence[TimeSeries],
    past_covariates: Optional[Sequence[TimeSeries]] = None,
    future_covariates: Optional[Sequence[TimeSeries]] = None,
    num_samples: int = 1,
    start: Optional[Union[pd.Timestamp, float, int]] = None,
    start_format: Literal["position", "value"] = "value",
    forecast_horizon: int = 1,
    stride: int = 1,
    overlap_end: bool = False,
    last_points_only: bool = True,
    show_warnings: bool = True,
    verbose: bool = False,
    predict_likelihood_parameters: bool = False,
    **kwargs,
) -> Union[Sequence[TimeSeries], Sequence[Sequence[TimeSeries]]]:
    """
    Optimized historical forecasts for TorchForecastingModels

    Rely on _check_optimizable_historical_forecasts() to check that the assumptions are verified.

    The data_transformers are applied in historical_forecasts (input and predictions)
    """
    bounds = []
    for idx, series_ in enumerate(series):
        past_covariates_ = past_covariates[idx] if past_covariates is not None else None
        future_covariates_ = (
            future_covariates[idx] if future_covariates is not None else None
        )
        # obtain forecastable indexes boundaries, adjust target & covariates boundaries accordingly
        (
            hist_fct_start,
            hist_fct_end,
            _,
            _,
            _,
            _,
            _,
            _,
        ) = _get_historical_forecast_boundaries(
            model=model,
            series=series_,
            series_idx=idx,
            past_covariates=past_covariates_,
            future_covariates=future_covariates_,
            start=start,
            start_format=start_format,
            forecast_horizon=forecast_horizon,
            overlap_end=overlap_end,
            stride=stride,
            freq=series_.freq,
            show_warnings=show_warnings,
        )
        left_bound = series_.get_index_at_point(hist_fct_start)
        # latest possible prediction start is one time step after end of target series
        if hist_fct_end > series_.end_time():
            right_bound = len(series_)
        else:
            right_bound = series_.get_index_at_point(hist_fct_end)
        bounds.append((left_bound, right_bound))

    bounds, cum_lengths = _process_predict_start_points_bounds(
        series=series,
        bounds=np.array(bounds),
        stride=stride,
    )

    # TODO: is there a better way to call the super().predict() from TorchForecastingModel, without having to
    #  import it? (avoid circular imports)
    tfm_cls = [
        cls
        for cls in model.__class__.__mro__
        if cls.__name__ == "TorchForecastingModel"
    ][0]
    super_predict_params = inspect.signature(super(tfm_cls, model).predict).parameters
    super(tfm_cls, model).predict(
        forecast_horizon,
        series,
        past_covariates,
        future_covariates,
        num_samples=num_samples,
        predict_likelihood_parameters=predict_likelihood_parameters,
        show_warnings=show_warnings,
        **{k: v for k, v in kwargs.items() if k in super_predict_params},
    )

    dataset = model._build_inference_dataset(
        target=series,
        n=forecast_horizon,
        past_covariates=past_covariates,
        future_covariates=future_covariates,
        stride=stride,
        bounds=bounds,
    )

    predictions = model.predict_from_dataset(
        forecast_horizon,
        dataset,
        verbose=verbose,
        num_samples=num_samples,
        predict_likelihood_parameters=predict_likelihood_parameters,
        **kwargs,
    )

    # torch models return list of time series in order of historical forecasts: we reorder per time series
    forecasts_list = []
    for series_idx in range(len(series)):
        pred_idx_start = 0 if not series_idx else cum_lengths[series_idx - 1]
        pred_idx_end = cum_lengths[series_idx]
        preds = predictions[pred_idx_start:pred_idx_end]
        if last_points_only:
            # torch predictions come with the entire horizon: we extract last values
            preds = TimeSeries.from_times_and_values(
                times=generate_index(
                    start=preds[0].end_time(),
                    length=len(preds),
                    freq=preds[0].freq * stride,
                ),
                values=np.concatenate(
                    [p.all_values(copy=False)[-1:, :, :] for p in preds], axis=0
                ),
                columns=preds[0].columns,
                static_covariates=preds[0].static_covariates,
                hierarchy=preds[0].hierarchy,
                metadata=preds[0].metadata,
            )
        forecasts_list.append(preds)
    return forecasts_list



"""
Anomaly Aggregators
-------------------

An anomaly aggregator can take multiple detected anomalies (in the form of binary TimeSeries, as coming from an anomaly
detector) and combine them into one. It can typically be used to combine the detections of multiple models into one
final detection.

The key method is `predict()`, which takes as input one (or multiple) multivariate binary TimeSeries where each
component represents the detection of a single model, and returns one (or multiple) univariate binary TimeSeries
representing the final detection.
"""

from darts.ad.aggregators.aggregators import Aggregator, FittableAggregator
from darts.ad.aggregators.and_aggregator import AndAggregator
from darts.ad.aggregators.ensemble_sklearn_aggregator import EnsembleSklearnAggregator
from darts.ad.aggregators.or_aggregator import OrAggregator

__all__ = [
    "Aggregator",
    "FittableAggregator",
    "AndAggregator",
    "EnsembleSklearnAggregator",
    "OrAggregator",
]



"""
Explainability
--------------
"""

from darts.explainability.explainability_result import (
    ShapExplainabilityResult,
    TFTExplainabilityResult,
    _ExplainabilityResult,
)
from darts.explainability.shap_explainer import ShapExplainer
from darts.logging import get_logger
from darts.utils.utils import NotImportedModule

logger = get_logger(__name__)
try:
    from darts.explainability.tft_explainer import TFTExplainer
except ModuleNotFoundError:
    logger.warning(
        "Support for Torch based explainers not available. "
        'To enable them, install "darts", "u8darts[torch]" or "u8darts[all]" (with pip); '
        'or "u8darts-torch" or "u8darts-all" (with conda).'
    )
    TFTExplainer = NotImportedModule(module_name="(Py)Torch", warn=False)

__all__ = [
    "ShapExplainabilityResult",
    "TFTExplainabilityResult",
    "_ExplainabilityResult",
    "ShapExplainer",
    "TFTExplainer",
]



from collections.abc import Sequence
from itertools import product

import numpy as np
import pytest

from darts import TimeSeries
from darts.ad.detectors.detectors import FittableDetector
from darts.ad.detectors.iqr_detector import IQRDetector
from darts.ad.detectors.quantile_detector import QuantileDetector
from darts.ad.detectors.threshold_detector import ThresholdDetector

list_Detector = [(ThresholdDetector, {"low_threshold": 0.2})]

list_FittableDetector = [(QuantileDetector, {"low_quantile": 0.2})]

list_detectors = list_Detector + list_FittableDetector

metric_func = ["accuracy", "recall", "f1", "precision"]

delta = 1e-05


class TestAnomalyDetectionDetector:
    np.random.seed(42)

    # univariate series
    np_train = np.random.normal(loc=10, scale=0.5, size=100)
    train = TimeSeries.from_values(np_train)

    np_test = np.random.normal(loc=10, scale=1, size=100)
    test = TimeSeries.from_times_and_values(train._time_index, np_test)

    np_anomalies = np.random.choice(a=[0, 1], size=100, p=[0.9, 0.1])
    anomalies = TimeSeries.from_times_and_values(train._time_index, np_anomalies)

    # multivariate series
    np_mts_train = np.random.normal(loc=[10, 5], scale=[0.5, 1], size=[100, 2])
    mts_train = TimeSeries.from_values(np_mts_train)

    np_mts_test = np.random.normal(loc=[10, 5], scale=[1, 1.5], size=[100, 2])
    mts_test = TimeSeries.from_times_and_values(mts_train._time_index, np_mts_test)

    np_mts_anomalies = np.random.choice(a=[0, 1], size=[100, 2], p=[0.9, 0.1])
    mts_anomalies = TimeSeries.from_times_and_values(
        mts_train._time_index, np_mts_anomalies
    )

    np_probabilistic = np.random.choice(a=[0, 1], p=[0.5, 0.5], size=[100, 1, 5])
    probabilistic = TimeSeries.from_values(np_probabilistic)

    @pytest.mark.parametrize(
        "detector_config,series",
        product(list_detectors, [(train, test), (mts_train, mts_test)]),
    )
    def test_detect_return_type(self, detector_config, series):
        """Check that detect() behave as expected"""
        detector_cls, detector_kwargs = detector_config
        ts_train, ts_test = series
        detector = detector_cls(**detector_kwargs)
        if isinstance(detector, FittableDetector):
            detector.fit(ts_train)

        # Check if return type is TimeSeries when input is a single series
        assert isinstance(detector.detect(ts_test), TimeSeries)
        # Check if return type is Sequence when input is a Sequence of series
        assert isinstance(detector.detect([ts_test]), Sequence)

        # Input cannot be probabilistic
        with pytest.raises(ValueError):
            detector.detect(self.probabilistic)

    @pytest.mark.parametrize("detector_config", list_detectors)
    def test_eval_metric_return_type(self, detector_config):
        """Check that eval_metric() behave as expected"""
        detector_cls, detector_kwargs = detector_config
        detector = detector_cls(**detector_kwargs)

        # univariate
        if isinstance(detector, FittableDetector):
            detector.fit(self.train)
        # Check if return type is float when input is a series
        assert isinstance(
            detector.eval_metric(anomalies=self.anomalies, pred_scores=self.test),
            float,
        )
        # Check if return type is Sequence when input is a Sequence of series
        assert isinstance(
            detector.eval_metric(anomalies=self.anomalies, pred_scores=[self.test]),
            Sequence,
        )

        # multivariate
        if isinstance(detector, FittableDetector):
            detector.fit(self.mts_train)
        # Check if return type is Sequence when input is a multivariate series
        assert isinstance(
            detector.eval_metric(
                anomalies=self.mts_anomalies, pred_scores=self.mts_test
            ),
            Sequence,
        )
        # Check if return type is Sequence when input is a multivariate series
        assert isinstance(
            detector.eval_metric(
                anomalies=self.mts_anomalies, pred_scores=[self.mts_test]
            ),
            Sequence,
        )

        # Input cannot be probabilistic
        with pytest.raises(ValueError):
            detector.eval_metric(
                anomalies=self.anomalies, pred_scores=self.probabilistic
            )

    @pytest.mark.parametrize(
        "config",
        [
            (
                ThresholdDetector,
                {"low_threshold": 4.8, "high_threshold": 10.5},
                {"low_threshold": [4.8, 4.8], "high_threshold": [10.5, 10.5]},
            ),
            (
                QuantileDetector,
                {"low_quantile": 0.05, "high_quantile": 0.95},
                {"low_quantile": [0.05, 0.05], "high_quantile": [0.95, 0.95]},
            ),
        ],
    )
    def test_bounded_detectors_parameters_broadcasting(self, config):
        """If two values are given for low and high, and a series of width 2 is given,
        then the results must be the same as a detector that was given only one value
        for low and high (will duplicate the value for each width)"""
        detector_cls, kwargs_1param, kwargs_2params = config

        # detector that should broadcast the parameters to match series' width
        detector = detector_cls(**kwargs_1param)
        # detector created with a number of parameters matching the series' width
        detector_2param = detector_cls(**kwargs_2params)
        if isinstance(detector, FittableDetector):
            detector.fit(self.mts_train)
            detector_2param.fit(self.mts_train)

        binary_detection = detector.detect(self.mts_test)
        binary_detection_2param = detector_2param.detect(self.mts_test)
        assert binary_detection == binary_detection_2param

    @pytest.mark.parametrize("detector_config", list_FittableDetector)
    def test_fit_detect_series_width(self, detector_config):
        detector_cls, detector_kwargs = detector_config
        detector = detector_cls(**detector_kwargs)

        # Need to call fit() before calling detect()
        with pytest.raises(ValueError):
            detector.detect(self.test)

        # Check if _fit_called is False
        assert not detector._fit_called

        with pytest.raises(ValueError):
            # fit on sequence with series that have different width
            detector.fit([self.train, self.mts_train])

        with pytest.raises(ValueError):
            # Input cannot be probabilistic
            detector.fit(self.probabilistic)

        # case1: fit on UTS
        detector1 = detector
        detector1.fit(self.train)

        # Check if _fit_called is True after being fitted
        assert detector1._fit_called

        with pytest.raises(ValueError):
            # series must be same width as series used for training
            detector1.detect(self.mts_test)

        # case2: fit on MTS
        detector2 = detector
        detector2.fit(self.mts_test)

        # Check if _fit_called is True after being fitted
        assert detector2._fit_called

        with pytest.raises(ValueError):
            # series must be same width as series used for training
            detector2.detect(self.train)

    def test_QuantileDetector_constructor(self):
        # Need to have at least one parameter (low, high) not None
        with pytest.raises(ValueError):
            QuantileDetector()
        with pytest.raises(ValueError):
            QuantileDetector(low_quantile=None, high_quantile=None)

        # Parameter low must be float or Sequence of float
        with pytest.raises(TypeError):
            QuantileDetector(low_quantile="0.5")
        with pytest.raises(TypeError):
            QuantileDetector(low_quantile=[0.2, "0.1"])

        # Parameter high must be float or Sequence of float
        with pytest.raises(TypeError):
            QuantileDetector(high_quantile="0.5")
        with pytest.raises(TypeError):
            QuantileDetector(high_quantile=[0.2, "0.1"])

        # if high and low are both sequences of length>1, they must be of the same size
        with pytest.raises(ValueError):
            QuantileDetector(low_quantile=[0.2, 0.1], high_quantile=[0.95, 0.8, 0.9])
        with pytest.raises(ValueError):
            QuantileDetector(low_quantile=[0.2, 0.1, 0.7], high_quantile=[0.95, 0.8])

        # Parameter must be between 0 and 1
        with pytest.raises(ValueError):
            QuantileDetector(high_quantile=1.1)
        with pytest.raises(ValueError):
            QuantileDetector(high_quantile=-0.2)
        with pytest.raises(ValueError):
            QuantileDetector(high_quantile=[-0.1, 0.9])
        with pytest.raises(ValueError):
            QuantileDetector(low_quantile=1.1)
        with pytest.raises(ValueError):
            QuantileDetector(low_quantile=-0.2)
        with pytest.raises(ValueError):
            QuantileDetector(low_quantile=[-0.2, 0.3])

        # Parameter high must be higher or equal than parameter low
        with pytest.raises(ValueError):
            QuantileDetector(low_quantile=0.7, high_quantile=0.2)
        with pytest.raises(ValueError):
            QuantileDetector(low_quantile=[0.2, 0.9], high_quantile=[0.95, 0.1])
        with pytest.raises(ValueError):
            QuantileDetector(low_quantile=0.2, high_quantile=[0.95, 0.1])
        with pytest.raises(ValueError):
            QuantileDetector(low_quantile=[0.2, 0.9], high_quantile=0.8)

        # Parameter high/low cannot be sequence of only None
        with pytest.raises(ValueError):
            QuantileDetector(low_quantile=[None, None, None])
        with pytest.raises(ValueError):
            QuantileDetector(high_quantile=[None, None, None])
        with pytest.raises(ValueError):
            QuantileDetector(low_quantile=[None], high_quantile=[None, None, None])

        # check that low_threshold and high_threshold are the same and no errors are raised
        detector = QuantileDetector(low_quantile=0.5, high_quantile=0.5)
        detector.fit(self.train)
        assert detector.low_threshold == detector.high_threshold

    @pytest.mark.parametrize(
        "detector_kwargs",
        [
            {"low_quantile": 0.1, "high_quantile": [0.8, 0.7]},
            {"low_quantile": [0.1, 0.2], "high_quantile": [0.8, 0.9]},
            {"low_quantile": [0.1, 0.2], "high_quantile": 0.8},
            {"low_quantile": [0.1, 0.2]},
            {"high_quantile": [0.1, 0.2]},
        ],
    )
    def test_quantile_detector_fit_detect_matching_width(self, detector_kwargs):
        """Widths of series should match the number of values given for high or/and low,
        if more than one value is provided for either of them.

        `self.train` series has only one component whereas model is created with 2 values for at
        least one of the model"""
        detector = QuantileDetector(**detector_kwargs)

        # during training
        with pytest.raises(ValueError):
            detector.fit(self.train)
        with pytest.raises(ValueError):
            detector.fit([self.train, self.mts_train])

        # during detection
        detector.fit(self.mts_train)
        with pytest.raises(ValueError):
            detector.detect(self.train)
        with pytest.raises(ValueError):
            detector.detect([self.train, self.mts_train])

    def test_ThresholdDetector_constructor(self):
        # Need to have at least one parameter (low, high) not None
        with pytest.raises(ValueError):
            ThresholdDetector()
        with pytest.raises(ValueError):
            ThresholdDetector(low_threshold=None, high_threshold=None)

        # if high and low are both sequences of length>1, they must be of the same size
        with pytest.raises(ValueError):
            ThresholdDetector(low_threshold=[0.2, 0.1], high_threshold=[0.95, 0.8, 0.9])
        with pytest.raises(ValueError):
            ThresholdDetector(low_threshold=[0.2, 0.1, 0.7], high_threshold=[0.95, 0.8])

        # Parameter high must be higher or equal than parameter low
        with pytest.raises(ValueError):
            ThresholdDetector(low_threshold=0.7, high_threshold=0.2)
        with pytest.raises(ValueError):
            ThresholdDetector(low_threshold=[0.2, 0.9], high_threshold=[0.95, 0.1])
        with pytest.raises(ValueError):
            ThresholdDetector(low_threshold=0.2, high_threshold=[0.95, 0.1])
        with pytest.raises(ValueError):
            ThresholdDetector(low_threshold=[0.2, 0.9], high_threshold=0.8)
        with pytest.raises(ValueError):
            ThresholdDetector(low_threshold=[0.2, 0.9, None], high_threshold=0.8)

        # Parameter high/low cannot be sequence of only None
        with pytest.raises(ValueError):
            ThresholdDetector(low_threshold=[None, None, None])
        with pytest.raises(ValueError):
            ThresholdDetector(high_threshold=[None, None, None])
        with pytest.raises(ValueError):
            ThresholdDetector(low_threshold=[None], high_threshold=[None, None, None])

        # widths of series used for scoring must match the number of values given for high or/and low,
        # if high and low have a length higher than 1

        detector = ThresholdDetector(low_threshold=0.1, high_threshold=[0.8, 0.7])
        with pytest.raises(ValueError):
            detector.detect(self.train)
        with pytest.raises(ValueError):
            detector.detect([self.train, self.mts_train])

        detector = ThresholdDetector(
            low_threshold=[0.1, 0.2], high_threshold=[0.8, 0.9]
        )
        with pytest.raises(ValueError):
            detector.detect(self.train)
        with pytest.raises(ValueError):
            detector.detect([self.train, self.mts_train])

        detector = ThresholdDetector(low_threshold=[0.1, 0.2], high_threshold=0.8)
        with pytest.raises(ValueError):
            detector.detect(self.train)
        with pytest.raises(ValueError):
            detector.detect([self.train, self.mts_train])

        detector = ThresholdDetector(low_threshold=[0.1, 0.2])
        with pytest.raises(ValueError):
            detector.detect(self.train)
        with pytest.raises(ValueError):
            detector.detect([self.train, self.mts_train])

        detector = ThresholdDetector(high_threshold=[0.1, 0.2])
        with pytest.raises(ValueError):
            detector.detect(self.train)
        with pytest.raises(ValueError):
            detector.detect([self.train, self.mts_train])

    @pytest.mark.parametrize(
        "config",
        [
            (
                ThresholdDetector,
                {"low_threshold": 9.5, "high_threshold": 10.5},
                {
                    "anomalies": 58,
                    "accuracy": 0.41,
                    "recall": 0.40,
                    "f1": 0.06349,
                    "precision": 0.03448,
                },
                None,
            ),
            (
                QuantileDetector,
                {"low_quantile": 0.05, "high_quantile": 0.95},
                {
                    "anomalies": 42,
                    "accuracy": 0.57,
                    "recall": 0.40,
                    "f1": 0.08510,
                    "precision": 0.04761,
                },
                (9.13658, 10.74007),
            ),
            (
                IQRDetector,
                {},
                {
                    "anomalies": 28,
                    "accuracy": 0.69,
                    "recall": 0.2,
                    "f1": 0.060606,
                    "precision": 0.035714,
                },
                (8.9444, 10.95811),
            ),
            (
                IQRDetector,
                {"scale": 1},
                {
                    "anomalies": 47,
                    "accuracy": 0.52,
                    "recall": 0.4,
                    "f1": 0.07692,
                    "precision": 0.042553,
                },
                (9.19611, 10.70640),
            ),
        ],
    )
    def test_bounded_detector_eval_metric_univariate(self, config):
        """Verifying the performance of the bounded detectors on an univariate example"""
        detector_cls, detector_kwargs, expected_values, fitted_params = config
        detector = detector_cls(**detector_kwargs)
        if isinstance(detector, FittableDetector):
            detector.fit(self.train)
        binary_detection = detector.detect(self.test)

        # Return of .detect() must be binary
        np.testing.assert_array_equal(
            binary_detection.values(copy=False),
            binary_detection.values(copy=False).astype(bool),
        )

        # Return of .detect() must be same len as input
        assert len(binary_detection) == len(self.test)

        assert (
            binary_detection.sum(axis=0).all_values().flatten()[0]
            == expected_values["anomalies"]
        )

        for m_func in metric_func:
            assert (
                np.abs(
                    expected_values[m_func]
                    - detector.eval_metric(self.anomalies, self.test, metric=m_func),
                )
                < delta
            )

        # check the fitted parameters
        if isinstance(detector, QuantileDetector):
            assert np.abs(fitted_params[0] - detector.low_threshold[0]) < delta
            assert np.abs(fitted_params[1] - detector.high_threshold[0]) < delta

    @pytest.mark.parametrize(
        "config",
        [
            (
                ThresholdDetector,
                {"low_threshold": [4.8, 4.8], "high_threshold": [10.5, 10.5]},
                {
                    "anomalies": [28, 52],
                    "accuracy": (0.71, 0.48),
                    "recall": (0.45454, 0.5),
                    "f1": (0.25641, 0.16129),
                    "precision": (0.17857, 0.09615),
                },
            ),
            (
                ThresholdDetector,
                {"low_threshold": [10, None], "high_threshold": [None, 5]},
                {
                    "anomalies": [48, 43],
                    "accuracy": (0.51, 0.57),
                    "recall": (0.45454, 0.5),
                    "f1": (0.16949, 0.18867),
                    "precision": (0.10416, 0.11627),
                },
            ),
            (
                QuantileDetector,
                {"low_quantile": [0.05, 0.05], "high_quantile": [0.95, 0.95]},
                {
                    "anomalies": [37, 38],
                    "accuracy": (0.58, 0.58),
                    "recall": (0.27272, 0.3),
                    "f1": (0.125, 0.125),
                    "precision": (0.08108, 0.07894),
                },
            ),
            (
                QuantileDetector,
                {"low_quantile": [0.05, None], "high_quantile": [None, 0.95]},
                {
                    "anomalies": [20, 19],
                    "accuracy": (0.69, 0.75),
                    "recall": (0.0, 0.2),
                    "f1": (0.0, 0.13793),
                    "precision": (0.0, 0.10526),
                },
            ),
            (
                IQRDetector,
                {"scale": [0.5, np.inf]},
                {
                    "anomalies": [46, 0],
                    "accuracy": (0.51, 0.9),
                    "recall": (0.363636, 0.0),
                    "f1": (0.14035, 0.0),
                    "precision": (0.08695, 0.0),
                },
            ),
            (
                IQRDetector,
                {"scale": [np.inf, 0.77]},
                {
                    "anomalies": [0, 34],
                    "accuracy": (0.89, 0.62),
                    "recall": (0.0, 0.3),
                    "f1": (0.0, 0.136363),
                    "precision": (0.0, 0.08823),
                },
            ),
            (
                IQRDetector,
                {"scale": [0.5, 0.77]},
                {
                    "anomalies": [46, 34],
                    "accuracy": (0.51, 0.62),
                    "recall": (0.363636, 0.3),
                    "f1": (0.14035, 0.136363),
                    "precision": (0.08695, 0.08823),
                },
            ),
        ],
    )
    def test_bounded_detector_performance_multivariate(self, config):
        """
        TODO: improve these tests to introduce some correlation between actual and detected anomalies
        """
        detector_cls, detector_kwargs, expected_values = config
        detector = detector_cls(**detector_kwargs)

        if isinstance(detector, FittableDetector):
            detector.fit(self.mts_train)
        binary_detection = detector.detect(self.mts_test)

        # output must have the same width as the input
        expected_width = self.mts_test.n_components
        assert binary_detection.width == expected_width
        for m_func in metric_func:
            assert (
                len(
                    detector.eval_metric(
                        self.mts_anomalies, self.mts_test, metric=m_func
                    )
                )
                == expected_width
            )

        # check number of anomalies detected in the first component
        assert (
            binary_detection["0"].sum(axis=0).all_values().flatten()[0]
            == expected_values["anomalies"][0]
        )
        # check number of anomalies detected in the second component
        assert (
            binary_detection["1"].sum(axis=0).all_values().flatten()[0]
            == expected_values["anomalies"][1]
        )

        # check each metric on each component of the series
        for m_func in metric_func:
            metric_vals = detector.eval_metric(
                self.mts_anomalies, self.mts_test, metric=m_func
            )
            assert np.abs(expected_values[m_func][0] - metric_vals[0]) < delta
            assert np.abs(expected_values[m_func][1] - metric_vals[1]) < delta

    def test_fit_detect(self):
        """Calling fit() then detect() and fit_detect() should yield the same results"""
        detector1 = QuantileDetector(low_quantile=0.05, high_quantile=0.95)
        detector1.fit(self.train)
        prediction1 = detector1.detect(self.train)

        detector2 = QuantileDetector(low_quantile=0.05, high_quantile=0.95)
        prediction2 = detector2.fit_detect(self.train)

        assert prediction1 == prediction2

    def test_IQRDetector_constructor(self):
        # Numbers in `scale must be non-negative numbers
        with pytest.raises(ValueError):
            IQRDetector(scale=-1)
        with pytest.raises(ValueError):
            IQRDetector(scale=[-2])
        with pytest.raises(ValueError):
            IQRDetector(scale=[3, -4])
        with pytest.raises(ValueError):
            IQRDetector(scale="3")

        IQRDetector()
        IQRDetector(scale=1.2345)
        IQRDetector(scale=0)
        IQRDetector(scale=[1, 2, np.inf, 3, 0])

    def test_iqr_detector_fit_detect_matching_width(self):
        """Widths of series should match the number of values given for `scale`,
        if more than one value is provided.

        `self.train` series has only one component whereas model is created with 2/3 values"""
        detector = IQRDetector(scale=[1.5, 1.5])

        # during training
        with pytest.raises(ValueError):
            detector.fit(self.train)
        with pytest.raises(ValueError):
            detector.fit([self.train, self.mts_train])

        # during detection
        detector.fit(self.mts_train)
        with pytest.raises(ValueError):
            detector.detect(self.train)
        with pytest.raises(ValueError):
            detector.detect([self.train, self.mts_train])

        # single `scale` but fit to wrong widths
        detector = IQRDetector(scale=1.5)
        detector.fit(self.train)
        with pytest.raises(ValueError):
            detector.detect(self.mts_train)

        detector = IQRDetector(scale=1.5)
        detector.fit(self.mts_train)
        with pytest.raises(ValueError):
            detector.detect(self.train)

        detector = IQRDetector(scale=[1.5])
        detector.fit(self.mts_train)
        with pytest.raises(ValueError):
            detector.detect(self.train)

    # Test if the IQR detector is actually using the IQR algorithm
    def test_iqr_detector_fit_logic(self):
        # concatenate everything along the time axis
        np_series = self.train.all_values(copy=False)

        q1 = np.quantile(np_series, q=0.25)
        q3 = np.quantile(np_series, q=0.75)

        # With scale=0 it should detect only outside the IQR
        detector = IQRDetector(scale=0)
        detector.fit(self.train)

        assert np.abs(detector.detector.low_threshold - q1) < delta
        assert np.abs(detector.detector.high_threshold - q3) < delta

        # With larger scale it should add "padding" accordingly
        detector = IQRDetector(scale=0.5)
        detector.fit(self.train)

        assert detector.detector.low_threshold < q1
        assert detector.detector.high_threshold > q3

    def test_iqr_detector_detect_logic(self):
        np.random.seed(24)

        values = np.random.uniform(low=0, high=10, size=30)
        nice_ts = TimeSeries.from_values(values)

        np_series = nice_ts.all_values(copy=False)
        q1 = np.quantile(np_series, q=0.25)
        q3 = np.quantile(np_series, q=0.75)

        diff = q3 - q1
        scale = 0.5
        expected_low_threshold = q1 - diff * scale
        expected_high_threshold = q3 + diff * scale

        expected_anomalies = 10
        expected_not_anomalies = 20

        not_anomalies = np.random.uniform(
            low=expected_low_threshold + delta,
            high=expected_high_threshold - delta,
            size=expected_not_anomalies,
        )
        anomalies_high = np.random.uniform(
            low=expected_high_threshold + delta,
            high=expected_high_threshold + 10,
            size=expected_anomalies // 2,
        )
        anomalies_low = np.random.uniform(
            low=expected_low_threshold - 10,
            high=expected_low_threshold - delta,
            size=expected_anomalies // 2,
        )
        anomalous_arr = np.hstack((anomalies_high, not_anomalies, anomalies_low))
        anomalous_ts = TimeSeries.from_values(anomalous_arr)

        detector = IQRDetector(scale=scale)
        detector.fit(nice_ts)

        assert (
            np.abs(detector.detector.low_threshold[0] - expected_low_threshold) < delta
        )
        assert (
            np.abs(detector.detector.high_threshold[0] - expected_high_threshold)
            < delta
        )

        detection = detector.detect(anomalous_ts)

        assert detection.sum(axis=0).all_values().flatten()[0] == expected_anomalies



from collections.abc import Sequence
from itertools import product

import numpy as np
import pytest
from pyod.models.knn import KNN
from scipy.stats import cauchy, expon, gamma, laplace, norm, poisson

from darts import TimeSeries, metrics
from darts.ad.scorers import (
    CauchyNLLScorer,
    ExponentialNLLScorer,
    FittableAnomalyScorer,
    GammaNLLScorer,
    GaussianNLLScorer,
    KMeansScorer,
    LaplaceNLLScorer,
    PoissonNLLScorer,
    PyODScorer,
    WassersteinScorer,
)
from darts.ad.scorers import DifferenceScorer as Difference
from darts.ad.scorers import NormScorer as Norm
from darts.ad.scorers.scorers import NLLScorer
from darts.models import MovingAverageFilter
from darts.utils.timeseries_generation import linear_timeseries

list_NonFittableAnomalyScorer = [
    Norm(component_wise=False),
    Norm(component_wise=True),
    Difference(),
    GaussianNLLScorer(),
    ExponentialNLLScorer(),
    PoissonNLLScorer(),
    LaplaceNLLScorer(),
    CauchyNLLScorer(),
    GammaNLLScorer(),
]

list_FittableAnomalyScorer = [
    (PyODScorer, {"model": KNN(), "component_wise": False}),
    (KMeansScorer, {"component_wise": False}),
    (WassersteinScorer, {"window_agg": False, "component_wise": False}),
]

# (scorer_cls, values, distribution, distribution_kwargs, prob_density_func, prob_density_func)
list_NLLScorer = [
    (
        CauchyNLLScorer,
        [3, 2, 0.5, 0.9],
        np.random.standard_cauchy,
        {},
        cauchy.pdf,
        None,
    ),
    (
        ExponentialNLLScorer,
        [3, 0.1, 2, 0.01],
        np.random.exponential,
        {"scale": 2.0},
        expon.pdf,
        None,
    ),
    (
        GammaNLLScorer,
        [3, 0.1, 2, 0.5],
        np.random.gamma,
        {"shape": 2, "scale": 2},
        gamma.pdf,
        {"a": 2, "scale": 2},
    ),
    (
        GaussianNLLScorer,
        [3, 0.1, -2, 0.01],
        np.random.normal,
        {"loc": 0, "scale": 2},
        norm.pdf,
        None,
    ),
    (
        LaplaceNLLScorer,
        [3, 10, -2, 0.01],
        np.random.laplace,
        {"loc": 0, "scale": 2},
        laplace.pdf,
        None,
    ),
    (
        PoissonNLLScorer,
        [3, 2, 10, 1],
        np.random.poisson,
        {"lam": 1},
        poisson.pmf,
        {"mu": 1},
    ),
]

delta = 1e-05


class TestAnomalyDetectionScorer:
    np.random.seed(42)

    # univariate series
    np_train = np.random.normal(loc=10, scale=0.5, size=100)
    train = TimeSeries.from_values(np_train)

    np_test = np.random.normal(loc=10, scale=2, size=100)
    test = TimeSeries.from_times_and_values(train._time_index, np_test)

    np_anomalies = np.random.choice(a=[0, 1], size=100, p=[0.9, 0.1])
    anomalies = TimeSeries.from_times_and_values(train._time_index, np_anomalies)

    np_only_1_anomalies = np.random.choice(a=[0, 1], size=100, p=[0, 1])
    only_1_anomalies = TimeSeries.from_times_and_values(
        train._time_index, np_only_1_anomalies
    )

    np_only_0_anomalies = np.random.choice(a=[0, 1], size=100, p=[1, 0])
    only_0_anomalies = TimeSeries.from_times_and_values(
        train._time_index, np_only_0_anomalies
    )

    modified_train = MovingAverageFilter(window=10).filter(train)
    modified_test = MovingAverageFilter(window=10).filter(test)

    np_probabilistic = np.random.normal(loc=10, scale=2, size=[100, 1, 20])
    probabilistic = TimeSeries.from_times_and_values(
        train._time_index, np_probabilistic
    )

    # multivariate series
    np_mts_train = np.random.normal(loc=[10, 5], scale=[0.5, 1], size=[100, 2])
    mts_train = TimeSeries.from_values(np_mts_train)

    np_mts_test = np.random.normal(loc=[10, 5], scale=[1, 1.5], size=[100, 2])
    mts_test = TimeSeries.from_times_and_values(mts_train._time_index, np_mts_test)

    np_mts_anomalies = np.random.choice(a=[0, 1], size=[100, 2], p=[0.9, 0.1])
    mts_anomalies = TimeSeries.from_times_and_values(
        mts_train._time_index, np_mts_anomalies
    )

    modified_mts_train = MovingAverageFilter(window=10).filter(mts_train)
    modified_mts_test = MovingAverageFilter(window=10).filter(mts_test)

    np_mts_probabilistic = np.random.normal(
        loc=[[10], [5]], scale=[[1], [1.5]], size=[100, 2, 20]
    )
    mts_probabilistic = TimeSeries.from_times_and_values(
        mts_train._time_index, np_mts_probabilistic
    )

    @pytest.mark.parametrize("scorer", list_NonFittableAnomalyScorer)
    def test_score_from_pred_non_fittable_scorer(self, scorer):
        # NLLScorer require deterministic `series`
        if isinstance(scorer, NLLScorer):
            # series and pred_series are both deterministic
            with pytest.raises(ValueError):
                scorer.score_from_prediction(series=self.test, pred_series=self.test)
            # series is probabilistic, pred_series is deterministic
            with pytest.raises(ValueError):
                scorer.score_from_prediction(
                    series=self.probabilistic, pred_series=self.train
                )

            score = scorer.score_from_prediction(
                series=self.train, pred_series=self.probabilistic
            )
            assert isinstance(score, TimeSeries)
            assert score.all_values().shape == (len(self.train), 1, 1)
        else:
            # Check if return type is float when input is a series
            assert isinstance(
                scorer.score_from_prediction(self.test, self.modified_test), TimeSeries
            )

            # Check if return type is Sequence when input is a Sequence of series
            assert isinstance(
                scorer.score_from_prediction([self.test], [self.modified_test]),
                Sequence,
            )

            # Check if return type is Sequence when input is a multivariate series
            assert isinstance(
                scorer.score_from_prediction(self.mts_test, self.modified_mts_test),
                TimeSeries,
            )

            # Check if return type is Sequence when input is a multivariate series
            assert isinstance(
                scorer.score_from_prediction([self.mts_test], [self.modified_mts_test]),
                Sequence,
            )

    @pytest.mark.parametrize("scorer_config", list_FittableAnomalyScorer)
    def test_score_return_type(self, scorer_config):
        scorer_cls, scorer_kwargs = scorer_config
        scorer = scorer_cls(**scorer_kwargs)

        scorer.fit(self.train)
        # Check if return type is float when input is a series
        assert isinstance(scorer.score(self.test), TimeSeries)

        # Check if return type is Sequence when input is a sequence of series
        assert isinstance(scorer.score([self.test]), Sequence)

        scorer.fit(self.mts_train)
        # Check if return type is Sequence when input is a multivariate series
        assert isinstance(scorer.score(self.mts_test), TimeSeries)

        # Check if return type is Sequence when input is a sequence of multivariate series
        assert isinstance(scorer.score([self.mts_test]), Sequence)

        # Check return types for score_from_prediction()
        scorer.fit_from_prediction(self.train, self.modified_train)
        # Check if return type is float when input is a series
        assert isinstance(
            scorer.score_from_prediction(self.test, self.modified_test), TimeSeries
        )

        # Check if return type is Sequence when input is a Sequence of series
        assert isinstance(
            scorer.score_from_prediction([self.test], [self.modified_test]),
            Sequence,
        )

        scorer.fit_from_prediction(self.mts_train, self.modified_mts_train)
        # Check if return type is Sequence when input is a multivariate series
        assert isinstance(
            scorer.score_from_prediction(self.mts_test, self.modified_mts_test),
            TimeSeries,
        )

        # Check if return type is Sequence when input is a multivariate series
        assert isinstance(
            scorer.score_from_prediction([self.mts_test], [self.modified_mts_test]),
            Sequence,
        )

    def test_eval_metric_from_prediction_return_type(self):
        scorer = Norm(component_wise=False)
        # Check if return type is float when input is a series
        assert isinstance(
            scorer.eval_metric_from_prediction(
                self.anomalies, self.test, self.modified_test
            ),
            float,
        )
        # Check if return type is Sequence when input is a Sequence of series
        assert isinstance(
            scorer.eval_metric_from_prediction(
                self.anomalies, [self.test], self.modified_test
            ),
            Sequence,
        )
        # Check if return type is a float when input is a multivariate series and component_wise is set to `False`
        assert isinstance(
            scorer.eval_metric_from_prediction(
                self.anomalies, self.mts_test, self.modified_mts_test
            ),
            float,
        )
        # Check if return type is Sequence when input is a multivariate series and component_wise is set to `False`
        assert isinstance(
            scorer.eval_metric_from_prediction(
                self.anomalies, [self.mts_test], self.modified_mts_test
            ),
            Sequence,
        )

        scorer = Norm(component_wise=True)
        # Check if return type is a float when input is a multivariate series and component_wise is set to `True`
        assert isinstance(
            scorer.eval_metric_from_prediction(
                self.mts_anomalies, self.mts_test, self.modified_mts_test
            ),
            Sequence,
        )
        # Check if return type is Sequence when input is a multivariate series and component_wise is set to `True`
        assert isinstance(
            scorer.eval_metric_from_prediction(
                self.mts_anomalies, [self.mts_test], self.modified_mts_test
            ),
            Sequence,
        )

    @pytest.mark.parametrize("scorer_config", list_FittableAnomalyScorer)
    def test_eval_metric_fittable_scorer(self, scorer_config):
        scorer_cls, scorer_kwargs = scorer_config
        fittable_scorer = scorer_cls(**scorer_kwargs)
        fittable_scorer.fit(self.train)

        # if component_wise set to False, 'anomalies' must have widths of 1
        with pytest.raises(ValueError):
            fittable_scorer.eval_metric(anomalies=self.mts_anomalies, series=self.test)
        with pytest.raises(ValueError):
            fittable_scorer.eval_metric(
                anomalies=[self.anomalies, self.mts_anomalies],
                series=[self.test, self.test],
            )

        # 'metric' must be str and "AUC_ROC" or "AUC_PR"
        with pytest.raises(ValueError):
            fittable_scorer.eval_metric(
                anomalies=self.anomalies, series=self.test, metric=1
            )
        with pytest.raises(ValueError):
            fittable_scorer.eval_metric(
                anomalies=self.anomalies,
                series=self.test,
                metric="auc_roc",
            )
        with pytest.raises(TypeError):
            fittable_scorer.eval_metric(
                anomalies=self.anomalies,
                series=self.test,
                metric=["AUC_ROC"],
            )

        # 'anomalies' must be binary
        with pytest.raises(ValueError):
            fittable_scorer.eval_metric(anomalies=self.test, series=self.test)

        # 'anomalies' must contain anomalies (at least one)
        with pytest.raises(ValueError):
            fittable_scorer.eval_metric(
                anomalies=self.only_0_anomalies, series=self.test
            )

        # 'anomalies' cannot contain only anomalies
        with pytest.raises(ValueError):
            fittable_scorer.eval_metric(
                anomalies=self.only_1_anomalies, series=self.test
            )

        # 'anomalies' must match the number of series if length higher than 1
        with pytest.raises(ValueError):
            fittable_scorer.eval_metric(
                anomalies=[self.anomalies, self.anomalies],
                series=self.test,
            )
        with pytest.raises(ValueError):
            fittable_scorer.eval_metric(
                anomalies=[self.anomalies, self.anomalies],
                series=[self.test, self.test, self.test],
            )

        # 'anomalies' must have non empty intersection with 'series'
        with pytest.raises(ValueError):
            fittable_scorer.eval_metric(
                anomalies=self.anomalies[:20], series=self.test[30:]
            )
        with pytest.raises(ValueError):
            fittable_scorer.eval_metric(
                anomalies=[self.anomalies, self.anomalies[:20]],
                series=[self.test, self.test[40:]],
            )

    @pytest.mark.parametrize(
        "scorer", [Norm(component_wise=False), KMeansScorer(component_wise=False)]
    )
    def test_eval_metric_from_prediction(self, scorer):
        if isinstance(scorer, FittableAnomalyScorer):
            scorer.fit(self.train)

        # name must be of type str
        assert isinstance(scorer.__str__(), str)

        # 'metric' must be str and "AUC_ROC" or "AUC_PR"
        with pytest.raises(ValueError):
            scorer.eval_metric_from_prediction(
                anomalies=self.anomalies,
                series=self.test,
                pred_series=self.modified_test,
                metric=1,
            )
        with pytest.raises(ValueError):
            scorer.eval_metric_from_prediction(
                anomalies=self.anomalies,
                series=self.test,
                pred_series=self.modified_test,
                metric="auc_roc",
            )
        with pytest.raises(TypeError):
            scorer.eval_metric_from_prediction(
                anomalies=self.anomalies,
                series=self.test,
                pred_series=self.modified_test,
                metric=["AUC_ROC"],
            )

        # 'anomalies' must be binary
        with pytest.raises(ValueError):
            scorer.eval_metric_from_prediction(
                anomalies=self.test,
                series=self.test,
                pred_series=self.modified_test,
            )

        # 'anomalies' must contain anomalies (at least one)
        with pytest.raises(ValueError):
            scorer.eval_metric_from_prediction(
                anomalies=self.only_0_anomalies,
                series=self.test,
                pred_series=self.modified_test,
            )

        # 'anomalies' cannot contain only anomalies
        with pytest.raises(ValueError):
            scorer.eval_metric_from_prediction(
                anomalies=self.only_1_anomalies,
                series=self.test,
                pred_series=self.modified_test,
            )

        # 'anomalies' must match the number of series if length higher than 1
        with pytest.raises(ValueError):
            scorer.eval_metric_from_prediction(
                anomalies=[self.anomalies, self.anomalies],
                series=[self.test, self.test, self.test],
                pred_series=[
                    self.modified_test,
                    self.modified_test,
                    self.modified_test,
                ],
            )
        with pytest.raises(ValueError):
            scorer.eval_metric_from_prediction(
                anomalies=[self.anomalies, self.anomalies],
                series=self.test,
                pred_series=self.modified_test,
            )

        # 'anomalies' must have non empty intersection with 'series' and 'pred_series'
        with pytest.raises(ValueError):
            scorer.eval_metric_from_prediction(
                anomalies=self.anomalies[:20],
                series=self.test[30:],
                pred_series=self.modified_test[30:],
            )
        with pytest.raises(ValueError):
            scorer.eval_metric_from_prediction(
                anomalies=[self.anomalies, self.anomalies[:20]],
                series=[self.test, self.test[40:]],
                pred_series=[self.modified_test, self.modified_test[40:]],
            )

    @pytest.mark.parametrize("scorer", list_NonFittableAnomalyScorer)
    def test_NonFittableAnomalyScorer(self, scorer):
        # Check if trainable is False, being a NonFittableAnomalyScorer
        assert not scorer.is_trainable

        # checks for score_from_prediction()
        # input must be Timeseries or sequence of Timeseries
        with pytest.raises(ValueError):
            scorer.score_from_prediction(self.train, "str")
        with pytest.raises(ValueError):
            scorer.score_from_prediction(
                [self.train, self.train], [self.modified_train, "str"]
            )
        # score on sequence with series that have different width
        with pytest.raises(ValueError):
            scorer.score_from_prediction(self.train, self.modified_mts_train)
        # input sequences have different length
        with pytest.raises(ValueError):
            scorer.score_from_prediction(
                [self.train, self.train], [self.modified_train]
            )
        # two inputs must have a non zero intersection
        with pytest.raises(ValueError):
            scorer.score_from_prediction(self.train[:50], self.train[55:])
        # every pairwise element must have a non zero intersection
        with pytest.raises(ValueError):
            scorer.score_from_prediction(
                [self.train, self.train[:50]], [self.train, self.train[55:]]
            )

    @pytest.mark.parametrize("scorer_config", list_FittableAnomalyScorer)
    def test_FittableAnomalyScorer(self, scorer_config):
        scorer_cls, scorer_kwargs = scorer_config
        fittable_scorer = scorer_cls(**scorer_kwargs)

        # Need to call fit() before calling score()
        with pytest.raises(ValueError):
            fittable_scorer.score(self.test)

        # Need to call fit() before calling score_from_prediction()
        with pytest.raises(ValueError):
            fittable_scorer.score_from_prediction(self.test, self.modified_test)

        # Check if _fit_called is False
        assert not fittable_scorer._fit_called

        # fit on sequence with series that have different width
        with pytest.raises(ValueError):
            fittable_scorer.fit([self.train, self.mts_train])

        # fit on sequence with series that have different width
        with pytest.raises(ValueError):
            fittable_scorer.fit_from_prediction(
                [self.train, self.mts_train],
                [self.modified_train, self.modified_mts_train],
            )

        # checks for fit_from_prediction()
        # input must be Timeseries or sequence of Timeseries
        with pytest.raises(ValueError):
            fittable_scorer.score_from_prediction(self.train, "str")
        with pytest.raises(ValueError):
            fittable_scorer.score_from_prediction(
                [self.train, self.train], [self.modified_train, "str"]
            )
        # two inputs must have the same length
        with pytest.raises(ValueError):
            fittable_scorer.fit_from_prediction(
                [self.train, self.train], [self.modified_train]
            )
        # two inputs must have the same width
        with pytest.raises(ValueError):
            fittable_scorer.fit_from_prediction([self.train], [self.modified_mts_train])
        # every element must have the same width
        with pytest.raises(ValueError):
            fittable_scorer.fit_from_prediction(
                [self.train, self.mts_train],
                [self.modified_train, self.modified_mts_train],
            )
        # two inputs must have a non zero intersection
        with pytest.raises(ValueError):
            fittable_scorer.fit_from_prediction(self.train[:50], self.train[55:])
        # every pairwise element must have a non zero intersection
        with pytest.raises(ValueError):
            fittable_scorer.fit_from_prediction(
                [self.train, self.train[:50]], [self.train, self.train[55:]]
            )

        # checks for fit()
        # input must be Timeseries or sequence of Timeseries
        with pytest.raises(ValueError):
            fittable_scorer.fit("str")
        with pytest.raises(ValueError):
            fittable_scorer.fit([self.modified_train, "str"])

        # checks for score_from_prediction()
        fittable_scorer.fit_from_prediction(self.train, self.modified_train)
        # input must be Timeseries or sequence of Timeseries
        with pytest.raises(ValueError):
            fittable_scorer.score_from_prediction(self.train, "str")
        with pytest.raises(ValueError):
            fittable_scorer.score_from_prediction(
                [self.train, self.train], [self.modified_train, "str"]
            )
        # two inputs must have the same length
        with pytest.raises(ValueError):
            fittable_scorer.score_from_prediction(
                [self.train, self.train], [self.modified_train]
            )
        # two inputs must have the same width
        with pytest.raises(ValueError):
            fittable_scorer.score_from_prediction(
                [self.train], [self.modified_mts_train]
            )
        # every element must have the same width
        with pytest.raises(ValueError):
            fittable_scorer.score_from_prediction(
                [self.train, self.mts_train],
                [self.modified_train, self.modified_mts_train],
            )
        # two inputs must have a non zero intersection
        with pytest.raises(ValueError):
            fittable_scorer.score_from_prediction(self.train[:50], self.train[55:])
        # every pairwise element must have a non zero intersection
        with pytest.raises(ValueError):
            fittable_scorer.score_from_prediction(
                [self.train, self.train[:50]], [self.train, self.train[55:]]
            )

        # checks for score()
        # input must be Timeseries or sequence of Timeseries
        with pytest.raises(ValueError):
            fittable_scorer.score("str")
        with pytest.raises(ValueError):
            fittable_scorer.score([self.modified_train, "str"])

        # caseA: fit with fit()
        # case1: fit on UTS
        fittable_scorerA1 = fittable_scorer
        fittable_scorerA1.fit(self.train)
        # Check if _fit_called is True after being fitted
        assert fittable_scorerA1._fit_called
        with pytest.raises(ValueError):
            # series must be same width as series used for training
            fittable_scorerA1.score(self.mts_test)
        # case2: fit on MTS
        fittable_scorerA2 = fittable_scorer
        fittable_scorerA2.fit(self.mts_train)
        # Check if _fit_called is True after being fitted
        assert fittable_scorerA2._fit_called
        with pytest.raises(ValueError):
            # series must be same width as series used for training
            fittable_scorerA2.score(self.test)

        # caseB: fit with fit_from_prediction()
        # case1: fit on UTS
        fittable_scorerB1 = fittable_scorer
        fittable_scorerB1.fit_from_prediction(self.train, self.modified_train)
        # Check if _fit_called is True after being fitted
        assert fittable_scorerB1._fit_called
        with pytest.raises(ValueError):
            # series must be same width as series used for training
            fittable_scorerB1.score_from_prediction(
                self.mts_test, self.modified_mts_test
            )
        # case2: fit on MTS
        fittable_scorerB2 = fittable_scorer
        fittable_scorerB2.fit_from_prediction(self.mts_train, self.modified_mts_train)
        # Check if _fit_called is True after being fitted
        assert fittable_scorerB2._fit_called
        with pytest.raises(ValueError):
            # series must be same width as series used for training
            fittable_scorerB2.score_from_prediction(self.test, self.modified_test)

    def test_Norm(self):
        # Check parameters
        self.expects_deterministic_input(Norm)

        # if component_wise=False must always return a univariate anomaly score
        scorer = Norm(component_wise=False)
        assert scorer.score_from_prediction(self.test, self.modified_test).width == 1

        assert (
            scorer.score_from_prediction(self.mts_test, self.modified_mts_test).width
            == 1
        )

        # if component_wise=True must always return the same width as the input
        scorer = Norm(component_wise=True)
        assert scorer.score_from_prediction(self.test, self.modified_test).width == 1
        assert (
            scorer.score_from_prediction(self.mts_test, self.modified_mts_test).width
            == self.mts_test.width
        )

        scorer = Norm(component_wise=True)
        # univariate case (equivalent to abs diff)
        assert scorer.score_from_prediction(self.test, self.test + 1).sum(
            axis=0
        ).all_values().flatten()[0] == len(self.test)
        assert scorer.score_from_prediction(self.test + 1, self.test).sum(
            axis=0
        ).all_values().flatten()[0] == len(self.test)

        # multivariate case with component_wise set to True (equivalent to abs diff)
        # abs(a - 2a) =  a
        assert (
            scorer.score_from_prediction(self.mts_test, self.mts_test * 2)["0"]
            == self.mts_test["0"]
        )
        assert (
            scorer.score_from_prediction(self.mts_test, self.mts_test * 2)["1"]
            == self.mts_test["1"]
        )

        # abs(2a - a) =  a
        assert (
            scorer.score_from_prediction(self.mts_test * 2, self.mts_test)["0"]
            == self.mts_test["0"]
        )
        assert (
            scorer.score_from_prediction(self.mts_test * 2, self.mts_test)["1"]
            == self.mts_test["1"]
        )

        scorer = Norm(component_wise=False)

        # univariate case (equivalent to abs diff)
        assert scorer.score_from_prediction(self.test, self.test + 1).sum(
            axis=0
        ).all_values().flatten()[0] == len(self.test)
        assert scorer.score_from_prediction(self.test + 1, self.test).sum(
            axis=0
        ).all_values().flatten()[0] == len(self.test)

        # multivariate case with component_wise set to False
        # norm(a - a + sqrt(2)) = 2 * len(a) with a being series of dim=2
        assert (
            np.abs(
                2 * len(self.mts_test)
                - scorer.score_from_prediction(
                    self.mts_test, self.mts_test + np.sqrt(2)
                )
                .sum(axis=0)
                .all_values()
                .flatten()[0]
            )
            < delta
        )

        assert not scorer.is_probabilistic

    def test_Difference(self):
        self.expects_deterministic_input(Difference)

        scorer = Difference()

        # univariate case
        assert scorer.score_from_prediction(self.test, self.test + 1).sum(
            axis=0
        ).all_values().flatten()[0] == -len(self.test)

        assert (
            scorer.score_from_prediction(self.test + 1, self.test)
            .sum(axis=0)
            .all_values()
            .flatten()[0]
        ) == len(self.test)

        # multivariate case
        # output of score() must be the same width as the width of the input
        assert (
            scorer.score_from_prediction(self.mts_test, self.mts_test).width
        ) == self.mts_test.width

        # a - 2a = - a
        assert (
            scorer.score_from_prediction(self.mts_test, self.mts_test * 2)["0"]
            == -self.mts_test["0"]
        )
        assert (
            scorer.score_from_prediction(self.mts_test, self.mts_test * 2)["1"]
            == -self.mts_test["1"]
        )
        # 2a - a =  a
        assert (
            scorer.score_from_prediction(self.mts_test * 2, self.mts_test)["0"]
            == self.mts_test["0"]
        )
        assert (
            scorer.score_from_prediction(self.mts_test * 2, self.mts_test)["1"]
            == self.mts_test["1"]
        )

        assert not scorer.is_probabilistic

    @staticmethod
    def helper_check_type_window(scorer, **kwargs):
        # window must be non-negative
        with pytest.raises(ValueError):
            scorer(window=-1, **kwargs)
        # window must be different from 0
        with pytest.raises(ValueError):
            scorer(window=0, **kwargs)

    def helper_window_parameter(self, scorer_to_test, **kwargs):
        self.helper_check_type_window(scorer_to_test, **kwargs)

        if scorer_to_test(**kwargs).is_trainable:
            # window must be smaller than the input of score()
            scorer = scorer_to_test(window=len(self.train) + 1, **kwargs)
            with pytest.raises(ValueError):
                scorer.fit(self.train)

            scorer = scorer_to_test(window=len(self.train) - 20, **kwargs)
            scorer.fit(self.train)
            with pytest.raises(ValueError):
                scorer.score(self.test[: len(self.train) // 2])

        else:
            # case only NLL scorers for now

            scorer = scorer_to_test(window=101)
            # window must be smaller than the input of score_from_prediction()
            with pytest.raises(ValueError):
                scorer.score_from_prediction(
                    series=self.test, pred_series=self.probabilistic
                )  # len(self.test)=100

    def diff_fn_parameter(self, scorer, **kwargs):
        # must be one of Darts per time step metrics (e.g. ae, err, ...)
        with pytest.raises(ValueError):
            scorer(diff_fn="abs_diff", **kwargs)
        # absolute error / absolute difference
        s_tmp = scorer(diff_fn=metrics.ae, **kwargs)
        diffs = s_tmp._diff_series([self.train], [self.test])
        assert diffs == [abs(self.train - self.test)]
        # error / difference
        s_tmp = scorer(diff_fn=metrics.err, **kwargs)
        diffs = s_tmp._diff_series([self.train], [self.test])
        assert diffs == [self.train - self.test]

    def component_wise_parameter(self, scorer_to_test, **kwargs):
        # if component_wise=False must always return a univariate anomaly score
        scorer = scorer_to_test(component_wise=False, **kwargs)
        scorer.fit(self.train)
        assert scorer.score(self.test).width == 1
        scorer.fit(self.mts_train)
        assert scorer.score(self.mts_test).width == 1

        # if component_wise=True must always return the same width as the input
        scorer = scorer_to_test(component_wise=True, **kwargs)
        scorer.fit(self.train)
        assert scorer.score(self.test).width == 1
        scorer.fit(self.mts_train)
        assert scorer.score(self.mts_test).width == self.mts_test.width

    def check_diff_series(self, scorer, **kwargs):
        # test _diff_series() directly: parameter must by "abs_diff" or "diff"
        with pytest.raises(ValueError):
            s_tmp = scorer(**kwargs)
            s_tmp.diff_fn = "random"
            s_tmp._diff_series(self.train, self.test)

    def expects_deterministic_input(self, scorer, **kwargs):
        scorer = scorer(**kwargs)
        if scorer.is_trainable:
            scorer.fit(self.train)
            np.testing.assert_warns(scorer.score(self.probabilistic))

        # always expects a deterministic input
        np.testing.assert_warns(
            scorer.score_from_prediction(self.train, self.probabilistic)
        )
        np.testing.assert_warns(
            scorer.score_from_prediction(self.probabilistic, self.train)
        )

    def test_WassersteinScorer(self):
        # Check parameters and inputs
        self.component_wise_parameter(WassersteinScorer)
        self.helper_window_parameter(WassersteinScorer)
        self.diff_fn_parameter(WassersteinScorer)
        self.expects_deterministic_input(WassersteinScorer)

        # test plotting (just call the functions)
        scorer = WassersteinScorer(window=2, window_agg=False)
        scorer.fit(self.train)
        scorer.show_anomalies(self.test, self.anomalies)
        with pytest.raises(ValueError):
            # should fail for a sequence of series
            scorer.show_anomalies([self.test, self.test], self.anomalies)
        scorer.show_anomalies_from_prediction(
            series=self.test,
            pred_series=self.test + 1,
            anomalies=self.anomalies,
        )
        with pytest.raises(ValueError):
            # should fail for a sequence of series
            scorer.show_anomalies_from_prediction(
                series=[self.test, self.test],
                pred_series=self.test + 1,
                anomalies=self.anomalies,
            )
        with pytest.raises(ValueError):
            # should fail for a sequence of series
            scorer.show_anomalies_from_prediction(
                series=self.test,
                pred_series=[self.test + 1, self.test + 2],
                anomalies=self.anomalies,
            )

        assert not scorer.is_probabilistic

    def test_univariate_Wasserstein(self):
        # univariate example
        np.random.seed(42)

        np_train_wasserstein = np.abs(np.random.normal(loc=0, scale=0.1, size=100))
        train_wasserstein = TimeSeries.from_times_and_values(
            self.train._time_index, np_train_wasserstein
        )

        np_test_wasserstein = np.abs(np.random.normal(loc=0, scale=0.1, size=100))
        np_first_anomaly = np.abs(np.random.normal(loc=0, scale=0.25, size=10))
        np_second_anomaly = np.abs(np.random.normal(loc=0.25, scale=0.05, size=5))
        np_third_anomaly = np.abs(np.random.normal(loc=0, scale=0.15, size=15))

        np_test_wasserstein[10:20] = np_first_anomaly
        np_test_wasserstein[40:45] = np_second_anomaly
        np_test_wasserstein[70:85] = np_third_anomaly
        test_wasserstein = TimeSeries.from_times_and_values(
            self.train._time_index, np_test_wasserstein
        )

        # create the anomaly series
        np_anomalies = np.zeros(len(test_wasserstein))
        np_anomalies[10:17] = 1
        np_anomalies[40:42] = 1
        np_anomalies[70:85] = 1
        anomalies_wasserstein = TimeSeries.from_times_and_values(
            test_wasserstein.time_index, np_anomalies, columns=["is_anomaly"]
        )

        # test model with window of 10
        scorer_10 = WassersteinScorer(window=10, window_agg=False)
        scorer_10.fit(train_wasserstein)
        auc_roc_w10 = scorer_10.eval_metric(
            anomalies_wasserstein, test_wasserstein, metric="AUC_ROC"
        )
        auc_pr_w10 = scorer_10.eval_metric(
            anomalies_wasserstein, test_wasserstein, metric="AUC_PR"
        )

        # test model with window of 20
        scorer_20 = WassersteinScorer(window=20, window_agg=False)
        scorer_20.fit(train_wasserstein)
        auc_roc_w20 = scorer_20.eval_metric(
            anomalies_wasserstein, test_wasserstein, metric="AUC_ROC"
        )
        auc_pr_w20 = scorer_20.eval_metric(
            anomalies_wasserstein, test_wasserstein, metric="AUC_PR"
        )

        assert np.abs(0.80637 - auc_roc_w10) < delta
        assert np.abs(0.83390 - auc_pr_w10) < delta
        assert np.abs(0.77828 - auc_roc_w20) < delta
        assert np.abs(0.93934 - auc_pr_w20) < delta

    def test_multivariate_componentwise_Wasserstein(self):
        # example multivariate WassersteinScorer component wise (True and False)
        np.random.seed(3)
        np_mts_train_wasserstein = np.abs(
            np.random.normal(loc=[0, 0], scale=[0.1, 0.2], size=[100, 2])
        )
        mts_train_wasserstein = TimeSeries.from_times_and_values(
            self.train._time_index, np_mts_train_wasserstein
        )

        np_mts_test_wasserstein = np.abs(
            np.random.normal(loc=[0, 0], scale=[0.1, 0.2], size=[100, 2])
        )
        np_first_anomaly_width1 = np.abs(np.random.normal(loc=0.5, scale=0.4, size=10))
        np_first_anomaly_width2 = np.abs(np.random.normal(loc=0, scale=0.5, size=10))
        np_first_commmon_anomaly = np.abs(
            np.random.normal(loc=0.5, scale=0.5, size=[10, 2])
        )

        np_mts_test_wasserstein[5:15, 0] = np_first_anomaly_width1
        np_mts_test_wasserstein[35:45, 1] = np_first_anomaly_width2
        np_mts_test_wasserstein[65:75, :] = np_first_commmon_anomaly

        mts_test_wasserstein = TimeSeries.from_times_and_values(
            mts_train_wasserstein._time_index, np_mts_test_wasserstein
        )

        # create the anomaly series width 1
        np_anomalies_width1 = np.zeros(len(mts_test_wasserstein))
        np_anomalies_width1[5:15] = 1
        np_anomalies_width1[65:75] = 1

        # create the anomaly series width 2
        np_anomaly_width2 = np.zeros(len(mts_test_wasserstein))
        np_anomaly_width2[35:45] = 1
        np_anomaly_width2[65:75] = 1

        anomalies_wasserstein_per_width = TimeSeries.from_times_and_values(
            mts_test_wasserstein.time_index,
            list(zip(*[np_anomalies_width1, np_anomaly_width2])),
            columns=["is_anomaly_0", "is_anomaly_1"],
        )

        # create the anomaly series for the entire series
        np_commmon_anomaly = np.zeros(len(mts_test_wasserstein))
        np_commmon_anomaly[5:15] = 1
        np_commmon_anomaly[35:45] = 1
        np_commmon_anomaly[65:75] = 1
        anomalies_common_wasserstein = TimeSeries.from_times_and_values(
            mts_test_wasserstein.time_index, np_commmon_anomaly, columns=["is_anomaly"]
        )

        # test scorer with component_wise=False
        scorer_w10_cwfalse = WassersteinScorer(
            window=10, component_wise=False, window_agg=False
        )
        scorer_w10_cwfalse.fit(mts_train_wasserstein)
        auc_roc_cwfalse = scorer_w10_cwfalse.eval_metric(
            anomalies_common_wasserstein, mts_test_wasserstein, metric="AUC_ROC"
        )

        # test scorer with component_wise=True
        scorer_w10_cwtrue = WassersteinScorer(
            window=10, component_wise=True, window_agg=False
        )
        scorer_w10_cwtrue.fit(mts_train_wasserstein)
        auc_roc_cwtrue = scorer_w10_cwtrue.eval_metric(
            anomalies_wasserstein_per_width, mts_test_wasserstein, metric="AUC_ROC"
        )

        assert np.abs(0.94637 - auc_roc_cwfalse) < delta
        assert np.abs(0.98606 - auc_roc_cwtrue[0]) < delta
        assert np.abs(0.96722 - auc_roc_cwtrue[1]) < delta

    def test_kmeansScorer(self):
        # Check parameters and inputs
        self.component_wise_parameter(KMeansScorer)
        self.helper_window_parameter(KMeansScorer)
        self.diff_fn_parameter(KMeansScorer)
        self.expects_deterministic_input(KMeansScorer)
        assert not KMeansScorer().is_probabilistic

    def test_univariate_kmeans(self):
        # univariate example
        np.random.seed(40)

        # create the train set
        np_width1 = np.random.choice(a=[0, 1], size=100, p=[0.5, 0.5])
        np_width2 = (np_width1 == 0).astype(float)
        KMeans_mts_train = TimeSeries.from_values(
            np.dstack((np_width1, np_width2))[0], columns=["component 1", "component 2"]
        )

        # create the test set
        # inject anomalies in the test timeseries
        np.random.seed(3)
        np_width1 = np.random.choice(a=[0, 1], size=100, p=[0.5, 0.5])
        np_width2 = (np_width1 == 0).astype(int)

        # 2 anomalies per type
        # type 1: random values for only one width
        np_width1[20:21] = 2
        np_width2[30:32] = 2

        # type 2: shift both widths values (+/- 1 for both widths)
        np_width1[45:47] = np_width1[45:47] + 1
        np_width2[45:47] = np_width2[45:47] + 1
        np_width1[60:64] = np_width1[65:69] - 1
        np_width2[60:64] = np_width2[65:69] - 1

        # type 3: switch one state to another for only one width (1 to 0 for one width)
        np_width1[75:82] = (np_width1[75:82] != 1).astype(int)
        np_width2[90:96] = (np_width2[90:96] != 1).astype(int)

        KMeans_mts_test = TimeSeries.from_values(
            np.dstack((np_width1, np_width2))[0], columns=["component 1", "component 2"]
        )

        # create the anomaly series
        anomalies_index = [
            20,
            30,
            31,
            45,
            46,
            60,
            61,
            62,
            63,
            75,
            76,
            77,
            78,
            79,
            80,
            81,
            90,
            91,
            92,
            93,
            94,
            95,
        ]
        np_anomalies = np.zeros(len(KMeans_mts_test))
        np_anomalies[anomalies_index] = 1
        KMeans_mts_anomalies = TimeSeries.from_times_and_values(
            KMeans_mts_test.time_index, np_anomalies, columns=["is_anomaly"]
        )

        kmeans_scorer = KMeansScorer(k=2, window=1, component_wise=False)
        kmeans_scorer.fit(KMeans_mts_train)

        metric_AUC_ROC = kmeans_scorer.eval_metric(
            KMeans_mts_anomalies, KMeans_mts_test, metric="AUC_ROC"
        )
        metric_AUC_PR = kmeans_scorer.eval_metric(
            KMeans_mts_anomalies, KMeans_mts_test, metric="AUC_PR"
        )

        assert metric_AUC_ROC == 1.0
        assert metric_AUC_PR == 1.0

    def test_multivariate_window_kmeans(self):
        # multivariate example with different windows
        np.random.seed(1)

        # create the train set
        np_series = np.zeros(100)
        np_series[0] = 2

        for i in range(1, len(np_series)):
            np_series[i] = np_series[i - 1] + np.random.choice(a=[-1, 1], p=[0.5, 0.5])
            if np_series[i] > 3:
                np_series[i] = 3
            if np_series[i] < 0:
                np_series[i] = 0

        ts_train = TimeSeries.from_values(np_series, columns=["series"])

        # create the test set
        np.random.seed(3)
        np_series = np.zeros(100)
        np_series[0] = 1

        for i in range(1, len(np_series)):
            np_series[i] = np_series[i - 1] + np.random.choice(a=[-1, 1], p=[0.5, 0.5])
            if np_series[i] > 3:
                np_series[i] = 3
            if np_series[i] < 0:
                np_series[i] = 0

        # 3 anomalies per type
        # type 1: sudden shift between state 0 to state 2 without passing by state 1
        np_series[23] = 3
        np_series[44] = 3
        np_series[91] = 0

        # type 2: having consecutive timestamps at state 1 or 2
        np_series[3:5] = 2
        np_series[17:19] = 1
        np_series[62:65] = 2

        ts_test = TimeSeries.from_values(np_series, columns=["series"])

        anomalies_index = [4, 23, 18, 44, 63, 64, 91]
        np_anomalies = np.zeros(100)
        np_anomalies[anomalies_index] = 1
        ts_anomalies = TimeSeries.from_times_and_values(
            ts_test.time_index, np_anomalies, columns=["is_anomaly"]
        )

        kmeans_scorer_w1 = KMeansScorer(k=4, window=1)
        kmeans_scorer_w1.fit(ts_train)

        kmeans_scorer_w2 = KMeansScorer(k=8, window=2, window_agg=False)
        kmeans_scorer_w2.fit(ts_train)

        auc_roc_w1 = kmeans_scorer_w1.eval_metric(
            ts_anomalies, ts_test, metric="AUC_ROC"
        )
        auc_pr_w1 = kmeans_scorer_w1.eval_metric(ts_anomalies, ts_test, metric="AUC_PR")

        auc_roc_w2 = kmeans_scorer_w2.eval_metric(
            ts_anomalies, ts_test, metric="AUC_ROC"
        )
        auc_pr_w2 = kmeans_scorer_w2.eval_metric(ts_anomalies, ts_test, metric="AUC_PR")

        assert np.abs(0.41551 - auc_roc_w1) < delta
        assert np.abs(0.064761 - auc_pr_w1) < delta
        assert np.abs(0.957513 - auc_roc_w2) < delta
        assert np.abs(0.88584 - auc_pr_w2) < delta

    def test_multivariate_componentwise_kmeans(self):
        # example multivariate KMeans component wise (True and False)
        np.random.seed(1)

        np_mts_train_kmeans = np.abs(
            np.random.normal(loc=[0, 0], scale=[0.1, 0.2], size=[100, 2])
        )
        mts_train_kmeans = TimeSeries.from_times_and_values(
            self.train._time_index, np_mts_train_kmeans
        )

        np_mts_test_kmeans = np.abs(
            np.random.normal(loc=[0, 0], scale=[0.1, 0.2], size=[100, 2])
        )
        np_first_anomaly_width1 = np.abs(np.random.normal(loc=0.5, scale=0.4, size=10))
        np_first_anomaly_width2 = np.abs(np.random.normal(loc=0, scale=0.5, size=10))
        np_first_commmon_anomaly = np.abs(
            np.random.normal(loc=0.5, scale=0.5, size=[10, 2])
        )

        np_mts_test_kmeans[5:15, 0] = np_first_anomaly_width1
        np_mts_test_kmeans[35:45, 1] = np_first_anomaly_width2
        np_mts_test_kmeans[65:75, :] = np_first_commmon_anomaly

        mts_test_kmeans = TimeSeries.from_times_and_values(
            mts_train_kmeans._time_index, np_mts_test_kmeans
        )

        # create the anomaly series width 1
        np_anomalies_width1 = np.zeros(len(mts_test_kmeans))
        np_anomalies_width1[5:15] = 1
        np_anomalies_width1[65:75] = 1

        # create the anomaly series width 2
        np_anomaly_width2 = np.zeros(len(mts_test_kmeans))
        np_anomaly_width2[35:45] = 1
        np_anomaly_width2[65:75] = 1

        anomalies_kmeans_per_width = TimeSeries.from_times_and_values(
            mts_test_kmeans.time_index,
            list(zip(*[np_anomalies_width1, np_anomaly_width2])),
            columns=["is_anomaly_0", "is_anomaly_1"],
        )

        # create the anomaly series for the entire series
        np_commmon_anomaly = np.zeros(len(mts_test_kmeans))
        np_commmon_anomaly[5:15] = 1
        np_commmon_anomaly[35:45] = 1
        np_commmon_anomaly[65:75] = 1
        anomalies_common_kmeans = TimeSeries.from_times_and_values(
            mts_test_kmeans.time_index, np_commmon_anomaly, columns=["is_anomaly"]
        )

        # test scorer with component_wise=False
        scorer_w10_cwfalse = KMeansScorer(
            window=10, component_wise=False, n_init=10, window_agg=False
        )
        scorer_w10_cwfalse.fit(mts_train_kmeans)
        auc_roc_cwfalse = scorer_w10_cwfalse.eval_metric(
            anomalies_common_kmeans, mts_test_kmeans, metric="AUC_ROC"
        )

        # test scorer with component_wise=True
        scorer_w10_cwtrue = KMeansScorer(
            window=10, component_wise=True, n_init=10, window_agg=False
        )
        scorer_w10_cwtrue.fit(mts_train_kmeans)
        auc_roc_cwtrue = scorer_w10_cwtrue.eval_metric(
            anomalies_kmeans_per_width, mts_test_kmeans, metric="AUC_ROC"
        )

        assert np.abs(1.0 - auc_roc_cwtrue[0]) < delta
        assert np.abs(0.97666 - auc_roc_cwtrue[1]) < delta
        assert np.abs(0.99007 - auc_roc_cwfalse) < delta

    def test_PyODScorer(self):
        # Check parameters and inputs
        self.component_wise_parameter(PyODScorer, model=KNN())
        self.helper_window_parameter(PyODScorer, model=KNN())
        self.diff_fn_parameter(PyODScorer, model=KNN())
        self.expects_deterministic_input(PyODScorer, model=KNN())
        assert not PyODScorer(model=KNN()).is_probabilistic

        # model parameter must be pyod.models type BaseDetector
        with pytest.raises(ValueError):
            PyODScorer(model=MovingAverageFilter(window=10))

        # component_wise parameter
        # if component_wise=False must always return a univariate anomaly score
        scorer = PyODScorer(model=KNN(), component_wise=False)
        scorer.fit(self.train)
        assert scorer.score(self.test).width == 1
        scorer.fit(self.mts_train)
        assert scorer.score(self.mts_test).width == 1
        # if component_wise=True must always return the same width as the input
        scorer = PyODScorer(model=KNN(), component_wise=True)
        scorer.fit(self.train)
        assert scorer.score(self.test).width == 1
        scorer.fit(self.mts_train)
        assert scorer.score(self.mts_test).width == self.mts_test.width

        # window parameter
        # window must be non-negative
        with pytest.raises(ValueError):
            PyODScorer(model=KNN(), window=-1)
        # window must be different from 0
        with pytest.raises(ValueError):
            PyODScorer(model=KNN(), window=0)

        # diff_fn parameter
        # must be None, 'diff' or 'abs_diff'
        with pytest.raises(ValueError):
            PyODScorer(model=KNN(), diff_fn="random")
        with pytest.raises(ValueError):
            PyODScorer(model=KNN(), diff_fn=1)

        scorer = PyODScorer(model=KNN())

        # model parameter must be pyod.models type BaseDetector
        with pytest.raises(ValueError):
            PyODScorer(model=MovingAverageFilter(window=10))

    def test_univariate_PyODScorer(self):
        # univariate test
        np.random.seed(40)

        # create the train set
        np_width1 = np.random.choice(a=[0, 1], size=100, p=[0.5, 0.5])
        np_width2 = (np_width1 == 0).astype(float)
        pyod_mts_train = TimeSeries.from_values(
            np.dstack((np_width1, np_width2))[0], columns=["component 1", "component 2"]
        )

        # create the test set
        # inject anomalies in the test timeseries
        np.random.seed(3)
        np_width1 = np.random.choice(a=[0, 1], size=100, p=[0.5, 0.5])
        np_width2 = (np_width1 == 0).astype(int)

        # 2 anomalies per type
        # type 1: random values for only one width
        np_width1[20:21] = 2
        np_width2[30:32] = 2

        # type 2: shift both widths values (+/- 1 for both widths)
        np_width1[45:47] = np_width1[45:47] + 1
        np_width2[45:47] = np_width2[45:47] + 1
        np_width1[60:64] = np_width1[65:69] - 1
        np_width2[60:64] = np_width2[65:69] - 1

        # type 3: switch one state to another for only one width (1 to 0 for one width)
        np_width1[75:82] = (np_width1[75:82] != 1).astype(int)
        np_width2[90:96] = (np_width2[90:96] != 1).astype(int)

        pyod_mts_test = TimeSeries.from_values(
            np.dstack((np_width1, np_width2))[0], columns=["component 1", "component 2"]
        )

        # create the anomaly series
        anomalies_index = [
            20,
            30,
            31,
            45,
            46,
            60,
            61,
            62,
            63,
            75,
            76,
            77,
            78,
            79,
            80,
            81,
            90,
            91,
            92,
            93,
            94,
            95,
        ]
        np_anomalies = np.zeros(len(pyod_mts_test))
        np_anomalies[anomalies_index] = 1
        pyod_mts_anomalies = TimeSeries.from_times_and_values(
            pyod_mts_test.time_index, np_anomalies, columns=["is_anomaly"]
        )

        pyod_scorer = PyODScorer(
            model=KNN(n_neighbors=10), component_wise=False, window=1
        )
        pyod_scorer.fit(pyod_mts_train)

        metric_AUC_ROC = pyod_scorer.eval_metric(
            pyod_mts_anomalies, pyod_mts_test, metric="AUC_ROC"
        )
        metric_AUC_PR = pyod_scorer.eval_metric(
            pyod_mts_anomalies, pyod_mts_test, metric="AUC_PR"
        )

        assert metric_AUC_ROC == 1.0
        assert metric_AUC_PR == 1.0

    def test_multivariate_window_PyODScorer(self):
        # multivariate example (with different window)
        np.random.seed(1)

        # create the train set
        np_series = np.zeros(100)
        np_series[0] = 2

        for i in range(1, len(np_series)):
            np_series[i] = np_series[i - 1] + np.random.choice(a=[-1, 1], p=[0.5, 0.5])
            if np_series[i] > 3:
                np_series[i] = 3
            if np_series[i] < 0:
                np_series[i] = 0

        ts_train = TimeSeries.from_values(np_series, columns=["series"])

        # create the test set
        np.random.seed(3)
        np_series = np.zeros(100)
        np_series[0] = 1

        for i in range(1, len(np_series)):
            np_series[i] = np_series[i - 1] + np.random.choice(a=[-1, 1], p=[0.5, 0.5])
            if np_series[i] > 3:
                np_series[i] = 3
            if np_series[i] < 0:
                np_series[i] = 0

        # 3 anomalies per type
        # type 1: sudden shift between state 0 to state 2 without passing by state 1
        np_series[23] = 3
        np_series[44] = 3
        np_series[91] = 0

        # type 2: having consecutive timestamps at state 1 or 2
        np_series[3:5] = 2
        np_series[17:19] = 1
        np_series[62:65] = 2

        ts_test = TimeSeries.from_values(np_series, columns=["series"])

        anomalies_index = [4, 23, 18, 44, 63, 64, 91]
        np_anomalies = np.zeros(100)
        np_anomalies[anomalies_index] = 1
        ts_anomalies = TimeSeries.from_times_and_values(
            ts_test.time_index, np_anomalies, columns=["is_anomaly"]
        )

        pyod_scorer_w1 = PyODScorer(
            model=KNN(n_neighbors=10), component_wise=False, window=1
        )
        pyod_scorer_w1.fit(ts_train)

        pyod_scorer_w2 = PyODScorer(
            model=KNN(n_neighbors=10),
            component_wise=False,
            window=2,
            window_agg=False,
        )
        pyod_scorer_w2.fit(ts_train)

        auc_roc_w1 = pyod_scorer_w1.eval_metric(ts_anomalies, ts_test, metric="AUC_ROC")
        auc_pr_w1 = pyod_scorer_w1.eval_metric(ts_anomalies, ts_test, metric="AUC_PR")

        auc_roc_w2 = pyod_scorer_w2.eval_metric(ts_anomalies, ts_test, metric="AUC_ROC")
        auc_pr_w2 = pyod_scorer_w2.eval_metric(ts_anomalies, ts_test, metric="AUC_PR")

        assert np.abs(0.5 - auc_roc_w1) < delta
        assert np.abs(0.07 - auc_pr_w1) < delta
        assert np.abs(0.957513 - auc_roc_w2) < delta
        assert np.abs(0.88584 - auc_pr_w2) < delta

    def test_multivariate_componentwise_PyODScorer(self):
        # multivariate example with component wise (True and False)
        np.random.seed(1)

        np_mts_train_PyOD = np.abs(
            np.random.normal(loc=[0, 0], scale=[0.1, 0.2], size=[100, 2])
        )
        mts_train_PyOD = TimeSeries.from_times_and_values(
            self.train._time_index, np_mts_train_PyOD
        )

        np_mts_test_PyOD = np.abs(
            np.random.normal(loc=[0, 0], scale=[0.1, 0.2], size=[100, 2])
        )
        np_first_anomaly_width1 = np.abs(np.random.normal(loc=0.5, scale=0.4, size=10))
        np_first_anomaly_width2 = np.abs(np.random.normal(loc=0, scale=0.5, size=10))
        np_first_commmon_anomaly = np.abs(
            np.random.normal(loc=0.5, scale=0.5, size=[10, 2])
        )

        np_mts_test_PyOD[5:15, 0] = np_first_anomaly_width1
        np_mts_test_PyOD[35:45, 1] = np_first_anomaly_width2
        np_mts_test_PyOD[65:75, :] = np_first_commmon_anomaly

        mts_test_PyOD = TimeSeries.from_times_and_values(
            mts_train_PyOD._time_index, np_mts_test_PyOD
        )

        # create the anomaly series width 1
        np_anomalies_width1 = np.zeros(len(mts_test_PyOD))
        np_anomalies_width1[5:15] = 1
        np_anomalies_width1[65:75] = 1

        # create the anomaly series width 2
        np_anomaly_width2 = np.zeros(len(mts_test_PyOD))
        np_anomaly_width2[35:45] = 1
        np_anomaly_width2[65:75] = 1

        anomalies_pyod_per_width = TimeSeries.from_times_and_values(
            mts_test_PyOD.time_index,
            list(zip(*[np_anomalies_width1, np_anomaly_width2])),
            columns=["is_anomaly_0", "is_anomaly_1"],
        )

        # create the anomaly series for the entire series
        np_commmon_anomaly = np.zeros(len(mts_test_PyOD))
        np_commmon_anomaly[5:15] = 1
        np_commmon_anomaly[35:45] = 1
        np_commmon_anomaly[65:75] = 1
        anomalies_common_PyOD = TimeSeries.from_times_and_values(
            mts_test_PyOD.time_index, np_commmon_anomaly, columns=["is_anomaly"]
        )

        # test scorer with component_wise=False
        scorer_w10_cwfalse = PyODScorer(
            model=KNN(n_neighbors=10),
            component_wise=False,
            window=10,
            window_agg=False,
        )
        scorer_w10_cwfalse.fit(mts_train_PyOD)
        auc_roc_cwfalse = scorer_w10_cwfalse.eval_metric(
            anomalies_common_PyOD, mts_test_PyOD, metric="AUC_ROC"
        )

        # test scorer with component_wise=True
        scorer_w10_cwtrue = PyODScorer(
            model=KNN(n_neighbors=10),
            component_wise=True,
            window=10,
            window_agg=False,
        )
        scorer_w10_cwtrue.fit(mts_train_PyOD)
        auc_roc_cwtrue = scorer_w10_cwtrue.eval_metric(
            anomalies_pyod_per_width, mts_test_PyOD, metric="AUC_ROC"
        )

        assert np.abs(0.990566 - auc_roc_cwfalse) < delta
        assert np.abs(1.0 - auc_roc_cwtrue[0]) < delta
        assert np.abs(0.98311 - auc_roc_cwtrue[1]) < delta

    @staticmethod
    def helper_evaluate_nll_scorer(
        NLLscorer_to_test,
        distribution_arrays,
        deterministic_values,
        real_NLL_values,
    ):
        NLLscorer_w1 = NLLscorer_to_test(window=1)
        NLLscorer_w2 = NLLscorer_to_test(window=2)

        assert NLLscorer_w1.is_probabilistic

        # create timeseries
        distribution_series = TimeSeries.from_values(
            distribution_arrays.reshape(2, 2, -1)
        )
        series = TimeSeries.from_values(
            np.array(deterministic_values).reshape(2, 2, -1)
        )

        # compute the NLL values with score_from_prediction for scorer with window=1 and 2
        # t -> timestamp, c -> component and w -> window used in scorer
        value_t1_c1_w1 = NLLscorer_w1.score_from_prediction(
            series[0]["0"], distribution_series[0]["0"]
        )
        value_t2_c1_w1 = NLLscorer_w1.score_from_prediction(
            series[1]["0"], distribution_series[1]["0"]
        )
        value_t1_2_c1_w1 = NLLscorer_w1.score_from_prediction(
            series["0"], distribution_series["0"]
        )
        value_t1_2_c1_w2 = NLLscorer_w2.score_from_prediction(
            series["0"], distribution_series["0"]
        )

        # check length
        assert len(value_t1_2_c1_w1) == 2
        # check width
        assert value_t1_2_c1_w1.width == 1

        # check equal value_test1 and value_test2
        assert value_t1_2_c1_w1[0] == value_t1_c1_w1
        assert value_t1_2_c1_w1[1] == value_t2_c1_w1

        # check if value_t1_2_c1_w1 is the - log likelihood
        np.testing.assert_array_almost_equal(
            # This is approximate because our NLL scorer is fit from samples
            value_t1_2_c1_w1.all_values().reshape(-1),
            real_NLL_values[::2],
            decimal=1,
        )

        # check if result is equal to avg of two values when window is equal to 2
        assert (
            value_t1_2_c1_w2.all_values().reshape(-1)[0]
            == value_t1_2_c1_w1.mean(axis=0).all_values().reshape(-1)[0]
        )

        # multivariate case
        # compute the NLL values with score_from_prediction for scorer with window=1 and window=2
        value_t1_2_c1_2_w1 = NLLscorer_w1.score_from_prediction(
            series, distribution_series
        )
        value_t1_2_c1_2_w2 = NLLscorer_w2.score_from_prediction(
            series, distribution_series
        )

        # check length
        assert len(value_t1_2_c1_2_w1) == 2
        assert len(value_t1_2_c1_2_w2) == 1
        # check width
        assert value_t1_2_c1_2_w1.width == 2
        assert value_t1_2_c1_2_w2.width == 2

        # check if value_t1_2_c1_2_w1 is the - log likelihood
        np.testing.assert_array_almost_equal(
            # This is approximate because our NLL scorer is fit from samples
            value_t1_2_c1_2_w1.all_values().reshape(-1),
            real_NLL_values,
            decimal=1,
        )

        # check if result is equal to avg of two values when window is equal to 2
        assert value_t1_2_c1_w2.all_values().reshape(-1) == value_t1_2_c1_w1.mean(
            axis=0
        ).all_values().reshape(-1)

    @pytest.mark.parametrize("config", list_NLLScorer)
    def test_nll_scorer(self, config):
        np.random.seed(4)

        (
            scorer_cls,
            values,
            distribution,
            dist_kwargs,
            prob_dens_func,
            pdf_kwargs,
        ) = config
        # some pdf don't have the same parameters as the corresponding distribution
        if pdf_kwargs is None:
            pdf_kwargs = dist_kwargs
        self.helper_window_parameter(scorer_cls)

        distribution = np.array([
            distribution(size=10000, **dist_kwargs) for _ in range(len(values))
        ])
        real_values = [-np.log(prob_dens_func(value, **pdf_kwargs)) for value in values]

        self.helper_evaluate_nll_scorer(scorer_cls, distribution, values, real_values)

    @pytest.mark.parametrize(
        "model,series",
        product(
            [(KMeansScorer, {"random_state": 42}), (PyODScorer, {"model": KNN()})],
            [(train, test), (mts_train, mts_test)],
        ),
    )
    def test_window_equal_one(self, model, series):
        """Check that model, created with window=1 generate the same score regardless of window_agg value."""
        ts_train, ts_test = series
        model_cls, model_kwargs = model

        scorer_T = model_cls(window=1, window_agg=True, **model_kwargs)
        scorer_F = model_cls(window=1, window_agg=False, **model_kwargs)

        scorer_T.fit(ts_train)
        scorer_F.fit(ts_train)

        auc_roc_T = scorer_T.eval_metric(anomalies=self.anomalies, series=ts_test)
        auc_roc_F = scorer_F.eval_metric(anomalies=self.anomalies, series=ts_test)

        assert auc_roc_T == auc_roc_F

    @pytest.mark.parametrize(
        "window,model,series",
        product(
            [2, 10, 39],
            [
                (KMeansScorer, {"random_state": 42}),
                (WassersteinScorer, {}),
                (PyODScorer, {"model": KNN()}),
            ],
            [(train, test), (mts_train, mts_test)],
        ),
    )
    def test_window_greater_than_one(self, window, model, series):
        """Check scorer with same window greater than 1 and different values of window_agg produce correct scores"""
        ts_train, ts_test = series
        model_cls, model_kwargs = model
        scorer_T = model_cls(window=window, window_agg=True, **model_kwargs)
        scorer_F = model_cls(window=window, window_agg=False, **model_kwargs)

        scorer_T.fit(ts_train)
        scorer_F.fit(ts_train)

        score_T = scorer_T.score(ts_test)
        score_F = scorer_F.score(ts_test)

        # same length
        assert len(score_T) == len(score_F)

        # same width
        assert score_T.width == score_F.width

        # same first time index
        assert score_T.time_index[0] == score_F.time_index[0]

        # same last time index
        assert score_T.time_index[-1] == score_F.time_index[-1]

        # same last value (by definition)
        assert score_T[-1] == score_F[-1]

    def test_fun_window_agg(self):
        """Verify that the anomaly score aggregation works as intended"""
        # window = 2, alternating anomaly scores
        window = 2
        scorer = KMeansScorer(window=window)
        anomaly_scores = TimeSeries.from_values(np.resize([1, -1], 10))
        aggreg_scores = scorer._fun_window_agg([anomaly_scores], window=window)[0]
        # in the last window, the score is not zeroed
        np.testing.assert_array_almost_equal(
            aggreg_scores.values(), np.array([[0, 0, 0, 0, 0, 0, 0, 0, 0, -1]]).T
        )
        assert aggreg_scores.time_index.equals(anomaly_scores.time_index)

        # window = 3, increment of 2 anomaly scores
        window = 3
        scorer = KMeansScorer(window=window)
        anomaly_scores = linear_timeseries(length=10, start_value=2, end_value=20)
        aggreg_scores = scorer._fun_window_agg([anomaly_scores], window=window)[0]
        # on the last "window" values, difference of only 1 between consecutive aggregated scores
        np.testing.assert_array_almost_equal(
            aggreg_scores.values(), np.array([[4, 6, 8, 10, 12, 14, 16, 18, 19, 20]]).T
        )
        assert aggreg_scores.time_index.equals(anomaly_scores.time_index)

        # window = 6, increment of 2 anomaly scores
        window = 6
        scorer = KMeansScorer(window=window)
        anomaly_scores = linear_timeseries(length=10, start_value=2, end_value=20)
        aggreg_scores = scorer._fun_window_agg([anomaly_scores], window=window)[0]
        # on the last "window" values, difference of only 1 between consecutive aggregated scores
        np.testing.assert_array_almost_equal(
            aggreg_scores.values(), np.array([[7, 9, 11, 13, 15, 16, 17, 18, 19, 20]]).T
        )
        assert aggreg_scores.time_index.equals(anomaly_scores.time_index)

        # window = 7, increment of 2 anomaly scores
        window = 7
        scorer = KMeansScorer(window=window)
        anomaly_scores = linear_timeseries(length=10, start_value=2, end_value=20)
        aggreg_scores = scorer._fun_window_agg([anomaly_scores], window=window)[0]
        # on the last "window" values, difference of only 1 between consecutive aggregated scores
        np.testing.assert_array_almost_equal(
            aggreg_scores.values(),
            np.array([[8, 10, 12, 14, 15, 16, 17, 18, 19, 20]]).T,
        )
        assert aggreg_scores.time_index.equals(anomaly_scores.time_index)



"""
Utils for time series generation
--------------------------------
"""

import math
from collections.abc import Sequence
from typing import Optional, Union

import holidays
import numpy as np
import pandas as pd

from darts import TimeSeries
from darts.logging import get_logger, raise_if, raise_if_not, raise_log
from darts.utils.utils import generate_index

logger = get_logger(__name__)

ONE_INDEXED_FREQS = {
    "day",
    "month",
    "quarter",
    "dayofyear",
    "day_of_year",
    "week",
    "weekofyear",
    "week_of_year",
}


def constant_timeseries(
    value: float = 1,
    start: Optional[Union[pd.Timestamp, int]] = pd.Timestamp("2000-01-01"),
    end: Optional[Union[pd.Timestamp, int]] = None,
    length: Optional[int] = None,
    freq: Union[str, int] = None,
    column_name: Optional[str] = "constant",
    dtype: np.dtype = np.float64,
) -> TimeSeries:
    """
    Creates a constant univariate TimeSeries with the given value, length (or end date), start date and frequency.

    Parameters
    ----------
    value
        The constant value that the TimeSeries object will assume at every index.
    start
        The start of the returned TimeSeries' index. If a pandas Timestamp is passed, the TimeSeries will have a pandas
        DatetimeIndex. If an integer is passed, the TimeSeries will have a pandas RangeIndex index. Works only with
        either `length` or `end`.
    end
        Optionally, the end of the returned index. Works only with either `start` or `length`. If `start` is
        set, `end` must be of same type as `start`. Else, it can be either a pandas Timestamp or an integer.
    length
        Optionally, the length of the returned index. Works only with either `start` or `end`.
    freq
        The time difference between two adjacent entries in the returned index. In case `start` is a timestamp,
        a DateOffset alias is expected; see
        `docs <https://pandas.pydata.org/pandas-docs/stable/user_guide/TimeSeries.html#dateoffset-objects>`_.
        By default, "D" (daily) is used.
        If `start` is an integer, `freq` will be interpreted as the step size in the underlying RangeIndex.
        The freq is optional for generating an integer index (if not specified, 1 is used).
    column_name
        Optionally, the name of the value column for the returned TimeSeries
    dtype
        The desired NumPy dtype (np.float32 or np.float64) for the resulting series

    Returns
    -------
    TimeSeries
        A constant TimeSeries with value 'value'.
    """

    index = generate_index(start=start, end=end, freq=freq, length=length)
    values = np.full(len(index), value, dtype=dtype)

    return TimeSeries.from_times_and_values(
        index, values, freq=freq, columns=pd.Index([column_name])
    )


def linear_timeseries(
    start_value: float = 0,
    end_value: float = 1,
    start: Optional[Union[pd.Timestamp, int]] = pd.Timestamp("2000-01-01"),
    end: Optional[Union[pd.Timestamp, int]] = None,
    length: Optional[int] = None,
    freq: Union[str, int] = None,
    column_name: Optional[str] = "linear",
    dtype: np.dtype = np.float64,
) -> TimeSeries:
    """
    Creates a univariate TimeSeries with a starting value of `start_value` that increases linearly such that
    it takes on the value `end_value` at the last entry of the TimeSeries. This means that
    the difference between two adjacent entries will be equal to
    (`end_value` - `start_value`) / (`length` - 1).

    Parameters
    ----------
    start_value
        The value of the first entry in the TimeSeries.
    end_value
        The value of the last entry in the TimeSeries.
    start
        The start of the returned TimeSeries' index. If a pandas Timestamp is passed, the TimeSeries will have a pandas
        DatetimeIndex. If an integer is passed, the TimeSeries will have a pandas RangeIndex index. Works only with
        either `length` or `end`.
    end
        Optionally, the end of the returned index. Works only with either `start` or `length`. If `start` is
        set, `end` must be of same type as `start`. Else, it can be either a pandas Timestamp or an integer.
    length
        Optionally, the length of the returned index. Works only with either `start` or `end`.
    freq
        The time difference between two adjacent entries in the returned index. In case `start` is a timestamp,
        a DateOffset alias is expected; see
        `docs <https://pandas.pydata.org/pandas-docs/stable/user_guide/TimeSeries.html#dateoffset-objects>`_.
        By default, "D" (daily) is used.
        If `start` is an integer, `freq` will be interpreted as the step size in the underlying RangeIndex.
        The freq is optional for generating an integer index (if not specified, 1 is used).
    column_name
        Optionally, the name of the value column for the returned TimeSeries
    dtype
        The desired NumPy dtype (np.float32 or np.float64) for the resulting series

    Returns
    -------
    TimeSeries
        A linear TimeSeries created as indicated above.
    """

    index = generate_index(start=start, end=end, freq=freq, length=length)
    values = np.linspace(start_value, end_value, len(index), dtype=dtype)
    return TimeSeries.from_times_and_values(
        index, values, freq=freq, columns=pd.Index([column_name])
    )


def sine_timeseries(
    value_frequency: float = 0.1,
    value_amplitude: float = 1.0,
    value_phase: float = 0.0,
    value_y_offset: float = 0.0,
    start: Optional[Union[pd.Timestamp, int]] = pd.Timestamp("2000-01-01"),
    end: Optional[Union[pd.Timestamp, int]] = None,
    length: Optional[int] = None,
    freq: Union[str, int] = None,
    column_name: Optional[str] = "sine",
    dtype: np.dtype = np.float64,
) -> TimeSeries:
    """
    Creates a univariate TimeSeries with a sinusoidal value progression with a given frequency, amplitude,
    phase and y offset.

    Parameters
    ----------
    value_frequency
        The number of periods that take place within one time unit given in `freq`.
    value_amplitude
        The maximum  difference between any value of the returned TimeSeries and `y_offset`.
    value_phase
        The relative position within one period of the first value of the returned TimeSeries (in radians).
    value_y_offset
        The shift of the sine function along the y axis.
    start
        The start of the returned TimeSeries' index. If a pandas Timestamp is passed, the TimeSeries will have a pandas
        DatetimeIndex. If an integer is passed, the TimeSeries will have a pandas RangeIndex index. Works only with
        either `length` or `end`.
    end
        Optionally, the end of the returned index. Works only with either `start` or `length`. If `start` is
        set, `end` must be of same type as `start`. Else, it can be either a pandas Timestamp or an integer.
    length
        Optionally, the length of the returned index. Works only with either `start` or `end`.
    freq
        The time difference between two adjacent entries in the returned index. In case `start` is a timestamp,
        a DateOffset alias is expected; see
        `docs <https://pandas.pydata.org/pandas-docs/stable/user_guide/TimeSeries.html#dateoffset-objects>`_.
        By default, "D" (daily) is used.
        If `start` is an integer, `freq` will be interpreted as the step size in the underlying RangeIndex.
        The freq is optional for generating an integer index (if not specified, 1 is used).
    column_name
        Optionally, the name of the value column for the returned TimeSeries
    dtype
        The desired NumPy dtype (np.float32 or np.float64) for the resulting series

    Returns
    -------
    TimeSeries
        A sinusoidal TimeSeries parametrized as indicated above.
    """

    index = generate_index(start=start, end=end, freq=freq, length=length)
    values = np.array(range(len(index)), dtype=dtype)
    f = np.vectorize(
        lambda x: value_amplitude
        * math.sin(2 * math.pi * value_frequency * x + value_phase)
        + value_y_offset
    )
    values = f(values)

    return TimeSeries.from_times_and_values(
        index, values, freq=freq, columns=pd.Index([column_name])
    )


def gaussian_timeseries(
    mean: Union[float, np.ndarray] = 0.0,
    std: Union[float, np.ndarray] = 1.0,
    start: Optional[Union[pd.Timestamp, int]] = pd.Timestamp("2000-01-01"),
    end: Optional[Union[pd.Timestamp, int]] = None,
    length: Optional[int] = None,
    freq: Union[str, int] = None,
    column_name: Optional[str] = "gaussian",
    dtype: np.dtype = np.float64,
) -> TimeSeries:
    """
    Creates a gaussian univariate TimeSeries by sampling all the series values independently,
    from a gaussian distribution with mean `mean` and standard deviation `std`.

    Parameters
    ----------
    mean
        The mean of the gaussian distribution that is sampled at each step.
        If a float value is given, the same mean is used at every step.
        If a numpy.ndarray of floats with the same length as `length` is
        given, a different mean is used at each time step.
    std
        The standard deviation of the gaussian distribution that is sampled at each step.
        If a float value is given, the same standard deviation is used at every step.
        If an array of dimension `(length, length)` is given, it will
        be used as covariance matrix for a multivariate gaussian distribution.
    start
        The start of the returned TimeSeries' index. If a pandas Timestamp is passed, the TimeSeries will have a pandas
        DatetimeIndex. If an integer is passed, the TimeSeries will have a pandas RangeIndex index. Works only with
        either `length` or `end`.
    end
        Optionally, the end of the returned index. Works only with either `start` or `length`. If `start` is
        set, `end` must be of same type as `start`. Else, it can be either a pandas Timestamp or an integer.
    length
        Optionally, the length of the returned index. Works only with either `start` or `end`.
    freq
        The time difference between two adjacent entries in the returned index. In case `start` is a timestamp,
        a DateOffset alias is expected; see
        `docs <https://pandas.pydata.org/pandas-docs/stable/user_guide/TimeSeries.html#dateoffset-objects>`_.
        By default, "D" (daily) is used.
        If `start` is an integer, `freq` will be interpreted as the step size in the underlying RangeIndex.
        The freq is optional for generating an integer index (if not specified, 1 is used).
    column_name
        Optionally, the name of the value column for the returned TimeSeries
    dtype
        The desired NumPy dtype (np.float32 or np.float64) for the resulting series

    Returns
    -------
    TimeSeries
        A white noise TimeSeries created as indicated above.
    """

    if isinstance(mean, np.ndarray):
        raise_if_not(
            mean.shape == (length,),
            "If a vector of means is provided, "
            "it requires the same length as the TimeSeries.",
            logger,
        )
    if isinstance(std, np.ndarray):
        raise_if_not(
            std.shape == (length, length),
            "If a matrix of standard deviations is provided, "
            "its shape has to match the length of the TimeSeries.",
            logger,
        )

    index = generate_index(start=start, end=end, freq=freq, length=length)
    values = np.random.normal(mean, std, size=len(index)).astype(dtype)

    return TimeSeries.from_times_and_values(
        index, values, freq=freq, columns=pd.Index([column_name])
    )


def random_walk_timeseries(
    mean: float = 0.0,
    std: float = 1.0,
    start: Optional[Union[pd.Timestamp, int]] = pd.Timestamp("2000-01-01"),
    end: Optional[Union[pd.Timestamp, int]] = None,
    length: Optional[int] = None,
    freq: Union[str, int] = None,
    column_name: Optional[str] = "random_walk",
    dtype: np.dtype = np.float64,
) -> TimeSeries:
    """
    Creates a random walk univariate TimeSeries, where each step is obtained by sampling a gaussian distribution
    with mean `mean` and standard deviation `std`.

    Parameters
    ----------
    mean
        The mean of the gaussian distribution that is sampled at each step.
    std
        The standard deviation of the gaussian distribution that is sampled at each step.
    start
        The start of the returned TimeSeries' index. If a pandas Timestamp is passed, the TimeSeries will have a pandas
        DatetimeIndex. If an integer is passed, the TimeSeries will have a pandas RangeIndex index. Works only with
        either `length` or `end`.
    end
        Optionally, the end of the returned index. Works only with either `start` or `length`. If `start` is
        set, `end` must be of same type as `start`. Else, it can be either a pandas Timestamp or an integer.
    length
        Optionally, the length of the returned index. Works only with either `start` or `end`.
    freq
        The time difference between two adjacent entries in the returned index. In case `start` is a timestamp,
        a DateOffset alias is expected; see
        `docs <https://pandas.pydata.org/pandas-docs/stable/user_guide/TimeSeries.html#dateoffset-objects>`_.
        By default, "D" (daily) is used.
        If `start` is an integer, `freq` will be interpreted as the step size in the underlying RangeIndex.
        The freq is optional for generating an integer index (if not specified, 1 is used).
    column_name
        Optionally, the name of the value column for the returned TimeSeries
    dtype
        The desired NumPy dtype (np.float32 or np.float64) for the resulting series

    Returns
    -------
    TimeSeries
        A random walk TimeSeries created as indicated above.
    """

    index = generate_index(start=start, end=end, freq=freq, length=length)
    values = np.cumsum(np.random.normal(mean, std, size=len(index)), dtype=dtype)

    return TimeSeries.from_times_and_values(
        index, values, freq=freq, columns=pd.Index([column_name])
    )


def autoregressive_timeseries(
    coef: Sequence[float],
    start_values: Optional[Sequence[float]] = None,
    start: Optional[Union[pd.Timestamp, int]] = pd.Timestamp("2000-01-01"),
    end: Optional[Union[pd.Timestamp, int]] = None,
    length: Optional[int] = None,
    freq: Union[str, int] = None,
    column_name: Optional[str] = "autoregressive",
) -> TimeSeries:
    """
    Creates a univariate, autoregressive TimeSeries whose values are calculated using specified coefficients `coef` and
    starting values `start_values`.

    Parameters
    ----------
    coef
        The autoregressive coefficients used for calculating the next time step.
        series[t] = coef[-1] * series[t-1] + coef[-2] * series[t-2] + ... + coef[0] * series[t-len(coef)]
    start_values
        The starting values used for calculating the first few values for which no lags exist yet.
        series[0] = coef[-1] * starting_values[-1] + coef[-2] * starting_values[-2] + ... + coef[0] * starting_values[0]
    start
        The start of the returned TimeSeries' index. If a pandas Timestamp is passed, the TimeSeries will have a pandas
        DatetimeIndex. If an integer is passed, the TimeSeries will have a pandas RangeIndex index. Works only with
        either `length` or `end`.
    end
        Optionally, the end of the returned index. Works only with either `start` or `length`. If `start` is
        set, `end` must be of same type as `start`. Else, it can be either a pandas Timestamp or an integer.
    length
        Optionally, the length of the returned index. Works only with either `start` or `end`.
    freq
        The time difference between two adjacent entries in the returned index. In case `start` is a timestamp,
        a DateOffset alias is expected; see
        `docs <https://pandas.pydata.org/pandas-docs/stable/user_guide/TimeSeries.html#dateoffset-objects>`_.
        By default, "D" (daily) is used.
        If `start` is an integer, `freq` will be interpreted as the step size in the underlying RangeIndex.
        The freq is optional for generating an integer index (if not specified, 1 is used).
    column_name
        Optionally, the name of the value column for the returned TimeSeries

    Returns
    -------
    TimeSeries
        An autoregressive TimeSeries created as indicated above.
    """

    # if no start values specified default to a list of 1s
    if start_values is None:
        start_values = np.ones(len(coef))
    else:
        raise_if_not(
            len(start_values) == len(coef),
            "start_values must have same length as coef.",
        )

    index = generate_index(start=start, end=end, freq=freq, length=length)

    values = np.empty(len(coef) + len(index))
    values[: len(coef)] = start_values

    for i in range(len(coef), len(coef) + len(index)):
        # calculate next time step as dot product of coefs with previous len(coef) time steps
        values[i] = np.dot(values[i - len(coef) : i], coef)

    return TimeSeries.from_times_and_values(
        index, values[len(coef) :], freq=freq, columns=pd.Index([column_name])
    )


def _extend_time_index_until(
    time_index: Union[pd.DatetimeIndex, pd.RangeIndex],
    until: Optional[Union[int, str, pd.Timestamp]],
    add_length: int,
) -> pd.DatetimeIndex:
    if not add_length and not until:
        return time_index

    raise_if(bool(add_length) and bool(until), "set only one of add_length and until")

    end = time_index[-1]
    freq = time_index.freq

    if add_length:
        raise_if_not(
            add_length >= 0,
            f"Expected add_length, by which to extend the time series by, "
            f"to be positive, got {add_length}",
        )

        try:
            end += add_length * freq
        except pd.errors.OutOfBoundsDatetime:
            raise_log(
                ValueError(
                    f"the add operation between {end} and {add_length * freq} will overflow"
                ),
                logger,
            )
    else:
        datetime_index = isinstance(time_index, pd.DatetimeIndex)

        if datetime_index:
            raise_if_not(
                isinstance(until, (str, pd.Timestamp)),
                "Expected valid timestamp for TimeSeries, "
                "indexed by DatetimeIndex, "
                f"for parameter until, got {type(end)}",
                logger,
            )
        else:
            raise_if_not(
                isinstance(until, int),
                "Expected integer for TimeSeries, indexed by RangeIndex, "
                f"for parameter until, got {type(end)}",
                logger,
            )

        timestamp = pd.Timestamp(until) if datetime_index else until

        raise_if_not(
            timestamp > end,
            f"Expected until, {timestamp} to lie past end of time index {end}",
        )

        ahead = timestamp - end
        raise_if_not(
            (ahead % freq) == pd.Timedelta(0),
            f"End date must correspond with frequency {freq} of the time axis",
            logger,
        )

        end = timestamp

    new_time_index = pd.date_range(start=time_index[0], end=end, freq=freq)
    return new_time_index


def holidays_timeseries(
    time_index: Union[TimeSeries, pd.DatetimeIndex],
    country_code: str,
    prov: str = None,
    state: str = None,
    column_name: Optional[str] = "holidays",
    until: Optional[Union[int, str, pd.Timestamp]] = None,
    add_length: int = 0,
    dtype: np.dtype = np.float64,
    tz: Optional[str] = None,
) -> TimeSeries:
    """
    Creates a binary univariate TimeSeries with index `time_index` that equals 1 at every index that lies within
    (or equals) a selected country's holiday, and 0 otherwise.

    Available countries can be found `here <https://github.com/dr-prodigy/python-holidays#available-countries>`_.

    Parameters
    ----------
    time_index
        Either a `pd.DatetimeIndex` or a `TimeSeries` for which to generate the holidays.
    country_code
        The country ISO code.
    prov
        The province.
    state
        The state.
    until
        Extend the time_index up until timestamp for datetime indexed series
        and int for range indexed series, should match or exceed forecasting window.
    add_length
        Extend the time_index by add_length, should match or exceed forecasting window.
        Set only one of until and add_length.
    column_name
        Optionally, the name of the value column for the returned TimeSeries.
    dtype
        The desired NumPy dtype (np.float32 or np.float64) for the resulting series.
    tz
        Optionally, a time zone to convert the time index to before generating the holidays.

    Returns
    -------
    TimeSeries
        A new binary holiday TimeSeries instance.
    """
    time_index_ts, time_index = _process_time_index(
        time_index=time_index,
        tz=tz,
        until=until,
        add_length=add_length,
    )

    scope = range(time_index[0].year, (time_index[-1] + pd.Timedelta(days=1)).year)
    country_holidays = holidays.country_holidays(
        country_code, prov=prov, state=state, years=scope
    )
    index_series = pd.Series(time_index, index=time_index)
    values = index_series.apply(lambda x: x in country_holidays).astype(dtype)
    return TimeSeries.from_times_and_values(
        time_index_ts, values, columns=pd.Index([column_name])
    )


def datetime_attribute_timeseries(
    time_index: Union[pd.DatetimeIndex, TimeSeries],
    attribute: str,
    one_hot: bool = False,
    cyclic: bool = False,
    until: Optional[Union[int, str, pd.Timestamp]] = None,
    add_length: int = 0,
    dtype=np.float64,
    with_columns: Optional[Union[list[str], str]] = None,
    tz: Optional[str] = None,
) -> TimeSeries:
    """
    Returns a new TimeSeries with index `time_index` and one or more dimensions containing
    (optionally one-hot encoded or cyclic encoded) pd.DatatimeIndex attribute information derived from the index.

    1-indexed attributes are shifted to enforce 0-indexing across all the encodings.

    Parameters
    ----------
    time_index
        Either a `pd.DatetimeIndex` attribute which will serve as the basis of the new column(s), or
        a `TimeSeries` whose time axis will serve this purpose.
    attribute
        An attribute of `pd.DatetimeIndex`, or `week` / `weekofyear` / `week_of_year` - e.g. "month", "weekday", "day",
        "hour", "minute", "second". See all available attributes in
        https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DatetimeIndex.html#pandas.DatetimeIndex.
    one_hot
        Boolean value indicating whether to add the specified attribute as a one hot encoding
        (results in more columns).
    cyclic
        Boolean value indicating whether to add the specified attribute as a cyclic encoding.
        Alternative to one_hot encoding, enable only one of the two.
        (adds 2 columns, corresponding to sin and cos transformation)
    until
        Extend the time_index up until timestamp for datetime indexed series
        and int for range indexed series, should match or exceed forecasting window.
    add_length
        Extend the time_index by add_length, should match or exceed forecasting window.
        Set only one of until and add_length.
    dtype
        The desired NumPy dtype (np.float32 or np.float64) for the resulting series
    with_columns
        Optionally, specify the output component names.
        * If `one_hot` and `cyclic` are ``False``, must be a string
        * If `cyclic` is ``True``, must be a list of two strings. The first string for the sine, the second for the
            cosine component name.
        * If `one_hot` is ``True``, must be a list of strings of the same length as the generated one hot encoded
            features.
    tz
        Optionally, a time zone to convert the time index to before computing the attributes.

    Returns
    -------
    TimeSeries
        New datetime attribute TimeSeries instance.
    """

    time_index_ts, time_index = _process_time_index(
        time_index=time_index,
        tz=tz,
        until=until,
        add_length=add_length,
    )

    raise_if_not(
        hasattr(pd.DatetimeIndex, attribute)
        or (attribute in ["week", "weekofyear", "week_of_year"]),
        f"attribute `{attribute}` needs to be an attribute of pd.DatetimeIndex. "
        "See all available attributes in "
        "https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DatetimeIndex.html#pandas.DatetimeIndex",
        logger,
    )

    raise_if(one_hot and cyclic, "set only one of one_hot or cyclic to true", logger)

    num_values_dict = {
        "month": 12,
        "day": 31,
        "weekday": 7,
        "dayofweek": 7,
        "day_of_week": 7,
        "hour": 24,
        "minute": 60,
        "second": 60,
        "microsecond": 1000000,
        "nanosecond": 1000,
        "quarter": 4,
        "dayofyear": 365,
        "day_of_year": 365,
        "week": 52,
        "weekofyear": 52,
        "week_of_year": 52,
    }

    if attribute not in ["week", "weekofyear", "week_of_year"]:
        values = getattr(time_index, attribute)
    else:
        values = (
            time_index.isocalendar()
            .set_index("week")
            .index.astype("int64")
            .rename("time")
        )

    # shift 1-indexed datetime attributes
    if attribute in ONE_INDEXED_FREQS:
        values -= 1

    # leap years insert an additional day on the 29th of February
    if attribute in {"dayofyear", "day_of_year"} and any(time_index.is_leap_year):
        num_values_dict[attribute] += 1

    # years contain an additional week if they are :
    # - a regular year starting on a thursday
    # - a leap year starting on a wednesday
    if attribute in {"week", "weekofyear", "week_of_year"}:
        years = time_index.year.unique()
        # check if year respect properties
        additional_week_year = any(
            ((not first_day.is_leap_year) and first_day.day_name() == "Thursday")
            or (first_day.is_leap_year and first_day.day_name() == "Wednesday")
            for first_day in [pd.Timestamp(f"{year}-01-01") for year in years]
        )
        # check if time index actually include the additional week
        additional_week_in_index = time_index[-1] - time_index[0] + pd.Timedelta(
            days=1
        ) >= pd.Timedelta(days=365)

        if additional_week_year and additional_week_in_index:
            num_values_dict[attribute] += 1

    if one_hot or cyclic:
        raise_if_not(
            attribute in num_values_dict,
            f"Given datetime attribute `{attribute}` not supported with one-hot or cyclical encoding. "
            f"Supported datetime attribute: {list(num_values_dict.keys())}",
            logger,
        )

    if one_hot:
        values_df = pd.get_dummies(values)
        # fill missing columns (in case not all values appear in time_index)
        attribute_range = np.arange(num_values_dict[attribute])
        is_missing = np.isin(attribute_range, values_df.columns.values, invert=True)
        # if there are attribute_range columns that are
        # not in values_df.columns.values
        if is_missing.any():
            dict_0 = {i: False for i in attribute_range[is_missing]}
            # Make a dataframe from the dictionary and concatenate it
            # to the values values_df  in which the existing columns
            values_df = pd.concat(
                [values_df, pd.DataFrame(dict_0, index=values_df.index)], axis=1
            ).sort_index(axis=1)
        else:
            values_df = values_df[attribute_range]

        if with_columns is None:
            with_columns = [
                attribute + "_" + str(column_name) for column_name in values_df.columns
            ]

        raise_if_not(
            len(with_columns) == len(values_df.columns),
            "For the given case with `one_hot=True`,`with_columns` must be a list of strings of length "
            f"{values_df.columns}.",
            logger=logger,
        )

        values_df.columns = with_columns
    else:
        if cyclic:
            if attribute == "day":
                periods = time_index.days_in_month.values
                freq = 2 * np.pi * np.reciprocal(periods.astype(dtype))
            else:
                period = num_values_dict[attribute]
                freq = 2 * np.pi / period

            if with_columns is None:
                with_columns = [attribute + "_sin", attribute + "_cos"]

            raise_if(
                len(with_columns) != 2,
                "`with_columns` must be a list of two strings when `cyclic=True`. "
                "The first string for the sine component name, the second for the cosine component name.",
                logger=logger,
            )
            values_df = pd.DataFrame({
                with_columns[0]: np.sin(freq * values),
                with_columns[1]: np.cos(freq * values),
            })
        else:
            if with_columns is None:
                with_columns = attribute
            raise_if_not(
                isinstance(with_columns, str),
                "`with_columns` must be a string specifying the output component name.",
                logger=logger,
            )
            values_df = pd.DataFrame({with_columns: values})

    values_df.index = time_index_ts
    return TimeSeries.from_dataframe(values_df).astype(dtype)


def _build_forecast_series(
    points_preds: Union[np.ndarray, Sequence[np.ndarray]],
    input_series: TimeSeries,
    custom_columns: list[str] = None,
    with_static_covs: bool = True,
    with_hierarchy: bool = True,
    pred_start: Optional[Union[pd.Timestamp, int]] = None,
    time_index: Union[pd.DatetimeIndex, pd.RangeIndex] = None,
) -> TimeSeries:
    """
    Builds a forecast time series starting after the end of an input time series, with the
    correct time index (or after the end of the input series, if specified).

    Parameters
    ----------
    points_preds
        Forecasted values, can be either the target(s) or parameters of the likelihood model
    input_series
        TimeSeries used as input for the prediction
    custom_columns
        New names for the forecast TimeSeries, used when the number of components changes
    with_static_covs
        If set to `False`, do not copy the input_series `static_covariates` attribute
    with_hierarchy
        If set to `False`, do not copy the input_series `hierarchy` attribute
    pred_start
        Optionally, give a custom prediction start point. Only effective if `time_index` is `None`.
    time_index
        Optionally, the index to use for the forecast time series.

    Returns
    -------
    TimeSeries
        New TimeSeries instance starting after the input series
    """
    if time_index is None:
        time_index_length = (
            len(points_preds)
            if isinstance(points_preds, np.ndarray)
            else len(points_preds[0])
        )
        time_index = _generate_new_dates(
            time_index_length,
            input_series=input_series,
            start=pred_start,
        )
    values = (
        points_preds
        if isinstance(points_preds, np.ndarray)
        else np.stack(points_preds, axis=2)
    )

    return TimeSeries.from_times_and_values(
        time_index,
        values,
        freq=input_series.freq_str,
        columns=input_series.columns if custom_columns is None else custom_columns,
        static_covariates=input_series.static_covariates if with_static_covs else None,
        hierarchy=input_series.hierarchy if with_hierarchy else None,
        metadata=input_series.metadata,
    )


def _generate_new_dates(
    n: int, input_series: TimeSeries, start: Optional[Union[pd.Timestamp, int]] = None
) -> Union[pd.DatetimeIndex, pd.RangeIndex]:
    """
    Generates `n` new dates after the end of the specified series
    """
    if start is None:
        last = input_series.end_time()
        start = last + input_series.freq
    return generate_index(
        start=start, freq=input_series.freq, length=n, name=input_series.time_dim
    )


def _process_time_index(
    time_index: Union[TimeSeries, pd.DatetimeIndex],
    tz: Optional[str] = None,
    until: Optional[Union[int, str, pd.Timestamp]] = None,
    add_length: int = 0,
) -> tuple[pd.DatetimeIndex, pd.DatetimeIndex]:
    """
    Extracts the time index, and optionally adds some time steps after the end of the index, and/or converts the
    index to another time zone.

    Returns a tuple of pd.DatetimeIndex with the first being the naive time index for generating a new TimeSeries,
    and the second being the one used for generating datetime attributes and holidays in a potentially different
    time zone.
    """
    if isinstance(time_index, TimeSeries):
        time_index = time_index.time_index

    if not isinstance(time_index, pd.DatetimeIndex):
        raise_log(
            ValueError(
                "`time_index` must be a pandas `DatetimeIndex` or a `TimeSeries` indexed with a `DatetimeIndex`."
            ),
            logger=logger,
        )
    if time_index.tz is not None:
        raise_log(
            ValueError("`time_index` must be time zone naive."),
            logger=logger,
        )
    time_index = _extend_time_index_until(time_index, until, add_length)

    # convert to another time zone
    if tz is not None:
        time_index_ = time_index.tz_localize("UTC").tz_convert(tz)
    else:
        time_index_ = time_index
    return time_index, time_index_



"""
AND Aggregator
--------------
"""

from collections.abc import Sequence

from darts import TimeSeries
from darts.ad.aggregators.aggregators import Aggregator
from darts.utils.utils import _parallel_apply


class AndAggregator(Aggregator):
    def __init__(self, n_jobs: int = 1) -> None:
        """AND Aggregator

        Aggregator that identifies a time step as anomalous if all the components are flagged as anomalous
        (logical AND).

        Parameters
        ----------
        n_jobs
            The number of jobs to run in parallel. Defaults to `1` (sequential). Setting the parameter to `-1` means
            using all the available processors.
        """
        super().__init__()
        self._n_jobs = n_jobs

    def __str__(self) -> str:
        return "AndAggregator"

    def _predict_core(
        self, series: Sequence[TimeSeries], *args, **kwargs
    ) -> Sequence[TimeSeries]:
        def _compononents_and(s: TimeSeries):
            return TimeSeries.from_times_and_values(
                times=s.time_index,
                values=(s.all_values(copy=False).sum(axis=1) >= s.width).astype(
                    s.dtype
                ),
                columns=["components_sum"],
            )

        return _parallel_apply(
            [(s,) for s in series],
            _compononents_and,
            n_jobs=1,
            fn_args=args,
            fn_kwargs=kwargs,
        )
